from lime.transitions import Line, label_decomposition, Particle
import numpy as np


def test_label_decomposition():

    # Components configuration
    fit_conf = {'O3_5007A_b': 'O3_5007A+O3_5007A_k-1+He1_5016A',
                'O3_5007A_m': 'O3_5007A+O3_5007A_k-1'}

    particle, wavelength, latex = label_decomposition('O3_5007A', fit_conf=fit_conf)
    assert particle[0] == 'O3'
    assert wavelength[0] == 5007.0
    assert latex[0] == r'$[OIII]5007\AA$'

    particle, wavelength, latex = label_decomposition('O3_5007A_b', fit_conf=fit_conf)
    assert particle[0] == 'O3'
    assert wavelength[0] == 5007.0
    assert latex[0] == r'$[OIII]5007\AA$'

    particle, wavelength, latex = label_decomposition('O3_5007A_m', fit_conf=fit_conf)
    assert particle[0] == 'O3'
    assert wavelength[0] == 5007.0
    assert latex[0] == r'$[OIII]5007\AA$+$[OIII]5007\AA-k_1$'

    return


class TestLineClass:

    def test_single_merged_blended(self):

        # Components configuration
        fit_conf = {'O3_5007A_b': 'O3_5007A+O3_5007A_k-1+He1_5016A',
                    'O3_5007A_m': 'O3_5007A+O3_5007A_k-1'}

        # Default band O3_5007A
        O3_band = np.array([4971.796688, 4984.514249, 4995.348943, 5024.303156, 5027.74326, 5043.797081])

        line = Line('O3_5007A', fit_conf=fit_conf)
        assert line.particle[0].label == 'O3'
        assert line.particle[0].symbol == 'O', line.particle[0].ionization == 3
        assert np.all(line.wavelength == np.array([5007.]))
        assert np.all(line.kinem == np.array([0]))
        assert np.all(line.transition_comp == np.array(['col']))
        assert np.all(line.profile_comp == np.array(['g-emi']))
        assert np.all(line.latex_label == np.array([r'$[OIII]5007\AA$']))
        assert np.all(line.list_comps == ['O3_5007A'])

        assert line.label == 'O3_5007A'
        assert line.profile_label is np.nan
        assert np.all(line.mask == O3_band)

        line = Line('O3_5007A_b', band=None, fit_conf=fit_conf)
        assert np.all(line.particle == [Particle.from_label('O3'), Particle.from_label('O3'), Particle.from_label('He1')])
        assert np.all(line.wavelength == np.array([5007., 5007., 5016.]))
        assert np.all(line.kinem == np.array([0, 1, 0]))
        assert np.all(line.transition_comp == np.array(['col', 'col', 'rec']))
        assert np.all(line.profile_comp == np.array(['g-emi', 'g-emi', 'g-emi']))
        assert np.all(line.latex_label == np.array([r'$[OIII]5007\AA$', r'$[OIII]5007\AA-k_1$', r'$HeI5016\AA$']))
        assert np.all(line.list_comps == ['O3_5007A', 'O3_5007A_k-1', 'He1_5016A'])

        assert line.label == 'O3_5007A_b'
        assert line.profile_label == 'O3_5007A+O3_5007A_k-1+He1_5016A'
        assert np.all(line.mask == O3_band)

        line = Line('O3_5007A_m', band=np.array([1, 2, 3, 4, 5, 6]), fit_conf=fit_conf)
        assert np.all(line.particle == [Particle('O3', symbol='O', ionization=3)])
        assert np.all(line.wavelength == np.array([5007]))
        assert np.all(line.kinem == np.array([0]))
        assert np.all(line.transition_comp == np.array(['col']))
        assert np.all(line.profile_comp == np.array(['g-emi', 'g-emi']))
        assert np.all(line.latex_label == np.array([r'$[OIII]5007\AA$+$[OIII]5007\AA-k_1$']))
        assert np.all(line.list_comps == ['O3_5007A_m'])

        assert line.label == 'O3_5007A_m'
        assert line.profile_label == 'O3_5007A+O3_5007A_k-1'
        assert np.all(line.mask == np.array([1, 2, 3, 4, 5, 6]))

        return

    def test_special_particles(self):

        line = Line('OIII_5007A')
        assert line.particle[0].label == 'OIII'
        assert np.all(line.wavelength == np.array([5007.]))
        assert np.all(line.kinem == np.array([0]))
        assert np.all(line.transition_comp == [None])
        assert np.all(line.profile_comp == np.array(['g-emi']))
        assert np.all(line.latex_label == np.array([r'OIII-$5007\AA$']))
        assert np.all(line.list_comps == ['OIII_5007A'])

        assert line.label == 'OIII_5007A'
        assert line.profile_label is np.nan
        assert line.mask is None

        line = Line('Halpha_6563A')
        assert line.particle[0].label == 'Halpha'
        assert np.all(line.wavelength == np.array([6563.]))
        assert np.all(line.kinem == np.array([0]))
        assert np.all(line.transition_comp == [None])
        assert np.all(line.profile_comp == np.array(['g-emi']))
        assert np.all(line.latex_label == np.array([r'Halpha-$6563\AA$']))
        assert np.all(line.list_comps == ['Halpha_6563A'])

        assert line.label == 'Halpha_6563A'
        assert line.profile_label is np.nan
        assert line.mask is None

        line = Line('HIPas4-3_18751A')
        assert line.particle[0].label == 'HIPas4-3'
        assert np.all(line.wavelength == np.array([18751.]))
        assert np.all(line.kinem == np.array([0]))
        assert np.all(line.transition_comp == [None])
        assert np.all(line.profile_comp == np.array(['g-emi']))
        assert np.all(line.latex_label == np.array([r'HIPas4-3-$18751\AA$']))
        assert np.all(line.list_comps == ['HIPas4-3_18751A'])

        assert line.label == 'HIPas4-3_18751A'
        assert line.profile_label is np.nan
        assert line.mask is None

        line = Line('OIII_5007A')
        assert line.particle[0].label == 'OIII'
        assert np.all(line.wavelength == np.array([5007.]))
        assert np.all(line.kinem == np.array([0]))
        assert np.all(line.transition_comp == [None])
        assert np.all(line.profile_comp == np.array(['g-emi']))
        assert np.all(line.latex_label == np.array([r'OIII-$5007\AA$']))
        assert np.all(line.list_comps == ['OIII_5007A'])

        assert line.label == 'OIII_5007A'
        assert line.profile_label is np.nan
        assert line.mask is None

        line = Line('C3_1909A')
        assert line.particle[0].label == 'C3'
        assert np.all(line.wavelength == np.array([1909.]))
        assert np.all(line.kinem == np.array([0]))
        assert np.all(line.transition_comp == ['sem'])
        assert np.all(line.profile_comp == np.array(['g-emi']))
        assert np.all(line.latex_label == np.array([r'$CIII]1909\AA$']))
        assert np.all(line.list_comps == ['C3_1909A'])

        assert line.label == 'C3_1909A'
        assert line.profile_label is np.nan
        assert line.mask is not None

        line = Line('C3_1909A_t-sem', band=None)
        assert line.particle[0].label == 'C3'
        assert np.all(line.wavelength == np.array([1909.]))
        assert np.all(line.kinem == np.array([0]))
        assert np.all(line.transition_comp == ['sem'])
        assert np.all(line.profile_comp == np.array(['g-emi']))
        assert np.all(line.latex_label == np.array([r'$CIII]1909\AA$']))
        assert np.all(line.list_comps == ['C3_1909A_t-sem'])

        assert line.label == 'C3_1909A_t-sem'
        assert line.profile_label is np.nan
        assert line.mask is None

        return


class TestParticleClass:

    def test_pyneb_items(self):

        particle = Particle.from_label('H1')

        assert particle.label == 'H1'
        assert particle.ionization == 1
        assert particle.symbol == 'H'

        particle2 = Particle('H1', 'H', ionization=1)
        assert particle2.label == 'H1'
        assert particle2.ionization == 1
        assert particle2.symbol == 'H'

        return



import numpy as np
import pandas as pd
import lime
from pathlib import Path
from lime.tools import int_to_roman, format_line_mask_option, refraction_index_air_vacuum

# Data for the tests
file_address = Path(__file__).parent/'data_tests'/'manga_spaxel.txt'
lines_log_address = Path(__file__).parent/'data_tests'/'manga_lines_log.txt'
lines_log = lime.load_log(lines_log_address)
redshift = 0.0475
norm_flux = 1e-17
wave_array, flux_array, err_array = np.loadtxt(file_address, unpack=True)
pixel_mask = np.isnan(err_array)

spec = lime.Spectrum(wave_array, flux_array, err_array, redshift=redshift, norm_flux=norm_flux,
                     pixel_mask=pixel_mask)

# Sample with 3 (repeated) observations
log_dict = {}
for i in range(3):
    log_dict[f'obj_{i}'] = lines_log.copy()

obj_list = [f'obj_{i}' for i in range(3)]
log_list = [lines_log_address] * 3
obs = lime.Sample.from_file_list(obj_list, log_list)


def test_int_to_roman():

    assert int_to_roman(1) == 'I'
    assert int_to_roman(10) == 'X'
    assert int_to_roman(1542) == 'MDXLII'

    return


def test_extract_fluxes():

    log1 = lines_log.copy()
    log2 = lines_log.copy()
    log3 = lines_log.copy()
    log4 = obs.log.copy()

    lime.extract_fluxes(log1)
    lime.extract_fluxes(log2, flux_type='intg', column_name='integrated', column_positions=3)
    lime.extract_fluxes(log3, flux_type='gauss', column_name='flux')
    lime.extract_fluxes(log4)

    # Mixture
    assert log1.loc['O3_5007A', 'line_flux'] == log1.loc['O3_5007A', 'gauss_flux']
    assert log1.loc['O3_4363A', 'line_flux'] == log1.loc['O3_4363A', 'intg_flux']
    assert log1.loc['O3_5007A', 'line_flux_err'] == log1.loc['O3_5007A', 'gauss_flux_err']
    assert log1.loc['O3_4363A', 'line_flux_err'] == log1.loc['O3_4363A', 'intg_flux_err']

    # Integrated
    assert log2.loc['O3_5007A', 'integrated'] == log2.loc['O3_5007A', 'intg_flux']
    assert log2.loc['O3_4363A', 'integrated'] == log2.loc['O3_4363A', 'intg_flux']
    assert log2.loc['O3_5007A', 'integrated_err'] == log2.loc['O3_5007A', 'intg_flux_err']
    assert log2.loc['O3_4363A', 'integrated_err'] == log2.loc['O3_4363A', 'intg_flux_err']

    # Gaussian
    assert log3.loc['O3_5007A', 'flux'] == log3.loc['O3_5007A', 'gauss_flux']
    assert log3.loc['O3_4363A', 'flux'] == log3.loc['O3_4363A', 'gauss_flux']
    assert log3.loc['O3_5007A', 'flux_err'] == log3.loc['O3_5007A', 'gauss_flux_err']
    assert log3.loc['O3_4363A', 'flux_err'] == log3.loc['O3_4363A', 'gauss_flux_err']

    # Multi - index
    assert np.all(log4.xs('O3_5007A', level='line')['line_flux'] == log4.xs('O3_5007A', level='line')['gauss_flux'])
    assert np.all(log4.xs('O3_4363A', level='line')['line_flux'] == log4.xs('O3_4363A', level='line')['intg_flux'])
    assert np.all(log4.xs('O3_5007A', level='line')['line_flux_err'] == log4.xs('O3_5007A', level='line')['gauss_flux_err'])
    assert np.all(log4.xs('O3_4363A', level='line')['line_flux_err'] == log4.xs('O3_4363A', level='line')['intg_flux_err'])

    # New columns position
    assert list(log1.columns).index('line_flux') == 0
    assert list(log1.columns).index('line_flux_err') == 1

    assert list(log2.columns).index('integrated') == 3
    assert list(log2.columns).index('integrated_err') == 4

    assert list(log3.columns).index('flux') == 0
    assert list(log3.columns).index('flux_err') == 1

    return


def test_extract_fluxes_single_index():

    log1 = lines_log.copy()
    log2 = lines_log.copy()
    log3 = lines_log.copy()

    # One line normalization
    lime.normalize_fluxes(log1, norm_list='H1_4861A')
    assert log1.loc['H1_6563A', 'line_flux'] == log1.loc['H1_6563A', 'gauss_flux']/log1.loc['H1_4861A', 'gauss_flux']
    assert log1.loc['O3_5007A', 'line_flux'] == log1.loc['O3_5007A', 'gauss_flux']/log1.loc['H1_4861A', 'gauss_flux']
    assert log1.loc['H1_4861A', 'line_flux'] == 1

    # List of normalizations
    norm_list = np.array(['H1_4861A'] * log2.index.size).astype(object)
    idcs_wide = log2.index.str.contains('k-1')
    norm_list[idcs_wide] = 'H1_4861A_k-1'
    lime.normalize_fluxes(log2, norm_list=norm_list)

    f_series = log2['gauss_flux']
    assert log2.loc['H1_6563A', 'line_flux'] == f_series['H1_6563A']/f_series['H1_4861A']
    assert log2.loc['O3_5007A', 'line_flux'] == f_series['O3_5007A']/f_series['H1_4861A']
    assert log2.loc['O3_5007A_k-1', 'line_flux'] == f_series['O3_5007A_k-1']/f_series['H1_4861A_k-1']
    assert log2.loc['O3_4959A_k-1', 'line_flux'] == f_series['O3_4959A_k-1']/f_series['H1_4861A_k-1']
    assert log2.loc['H1_4861A', 'line_flux'] == 1
    assert log2.loc['H1_4861A_k-1', 'line_flux'] == 1

    # Compound of normalizations
    line_list = ['H1_6563A/H1_4861A', 'O3_5007A/H1_4861A', 'O3_5007A_k-1/H1_4861A_k-1',
                 'O3_4959A_k-1/H1_4861A_k-1', 'H1_4861A_k-1/H1_4861A']
    lime.normalize_fluxes(log3, line_list=line_list)

    f_series = log3['gauss_flux']
    assert log3.loc['H1_6563A', 'line_flux'] == f_series['H1_6563A']/f_series['H1_4861A']
    assert log3.loc['O3_5007A', 'line_flux'] == f_series['O3_5007A']/f_series['H1_4861A']
    assert log3.loc['O3_5007A_k-1', 'line_flux'] == f_series['O3_5007A_k-1']/f_series['H1_4861A_k-1']
    assert log3.loc['O3_4959A_k-1', 'line_flux'] == f_series['O3_4959A_k-1']/f_series['H1_4861A_k-1']
    assert np.isnan(log3.loc['H1_4861A', 'line_flux'])
    assert log3.loc['H1_4861A_k-1', 'line_flux'] == f_series['H1_4861A_k-1']/f_series['H1_4861A']
    assert np.isnan(log3.loc['O3_4363A', 'line_flux'])

    return


def test_extract_fluxes_multi_index():

    idcs_remove = (obs.log.index.get_level_values('id') == 'obj_1') &\
                  (~obs.log.index.get_level_values('line').isin(['H1_6563A', 'O3_4363A', 'H1_4861A']))

    log1 = obs.log.loc[~idcs_remove].copy()
    log2 = obs.log.copy()
    log3 = obs.log.copy()
    log4 = obs.log.copy()

    # One line normalization
    lime.normalize_fluxes(log1, norm_list='H1_4861A')
    ratio = log1.xs('O3_5007A', level='line')['gauss_flux'] / log1.xs('H1_4861A', level='line')['gauss_flux']
    assert log1.xs('O3_5007A', level='line')['line_flux']['obj_0'] == ratio['obj_0']
    assert log1.xs('O3_5007A', level='line')['line_flux']['obj_2'] == ratio['obj_2']
    assert np.isnan(ratio['obj_1'])
    assert 'obj_1' not in log1.xs('O3_5007A', level='line')['line_flux']
    assert np.all(log1['norm_line'] == 'H1_4861A')

    # List of normalizations
    norm_list = np.array(['H1_4861A'] * log2.index.size).astype(object)
    idcs_wide = log2.index.get_level_values('line').str.contains('k-1')
    norm_list[idcs_wide] = 'H1_4861A_k-1'
    lime.normalize_fluxes(log2, norm_list=norm_list)

    ratio = log2.xs('O3_5007A', level='line')['gauss_flux'] / log2.xs('H1_4861A', level='line')['gauss_flux']
    assert log2.xs('O3_5007A', level='line')['line_flux']['obj_0'] == ratio['obj_0']
    assert log2.xs('O3_5007A', level='line')['line_flux']['obj_1'] == ratio['obj_1']
    assert log2.xs('O3_5007A', level='line')['line_flux']['obj_2'] == ratio['obj_2']

    assert np.all(log2.loc[~idcs_wide, 'norm_line'] == 'H1_4861A')
    assert np.all(log2.loc[idcs_wide, 'norm_line'] == 'H1_4861A_k-1')

    lime.normalize_fluxes(log3, line_list='H1_6563A', norm_list='H1_4861A')
    Halpha_norm = lines_log.loc['H1_6563A', 'gauss_flux'] / lines_log.loc['H1_4861A', 'gauss_flux']
    assert np.all(log3.xs('H1_6563A', level='line')['line_flux'] == Halpha_norm)
    assert pd.notnull(log3.line_flux).sum() == 3

    lime.normalize_fluxes(log4, line_list=['O3_5007A/H1_4861A', 'N2_6584A/H1_6563A'])
    O3_norm = lines_log.loc['O3_5007A', 'gauss_flux'] / lines_log.loc['H1_4861A', 'gauss_flux']
    N2_norm = lines_log.loc['N2_6584A', 'gauss_flux'] / lines_log.loc['H1_6563A', 'gauss_flux']

    assert np.all(log4.xs('O3_5007A', level='line')['line_flux'] == O3_norm)
    assert np.all(log4.xs('N2_6584A', level='line')['line_flux'] == N2_norm)

    assert np.all(log4.xs('O3_5007A', level='line')['norm_line'] == 'H1_4861A')
    assert np.all(log4.xs('N2_6584A', level='line')['norm_line'] == 'H1_6563A')

    return


def test_redshift_calculation():

    # Single index
    z_df = lime.redshift_calculation(lines_log)
    z_df_eqw = lime.redshift_calculation(lines_log, weight_parameter='eqw')
    z_df_flux_gauss = lime.redshift_calculation(lines_log, weight_parameter='gauss_flux')
    z_df_strong = lime.redshift_calculation(lines_log, line_list=['O3_5007A', 'H1_6563A'])

    assert np.allclose(z_df['z_mean'][0], 0.047526, atol=0.00024, equal_nan=True)
    assert np.allclose(z_df_eqw['z_mean'][0], 0.047526, atol=0.00024, equal_nan=True)
    assert np.allclose(z_df_flux_gauss['z_mean'][0], 0.047526, atol=0.00024, equal_nan=True)
    assert np.allclose(z_df_strong['z_mean'][0], 0.047498, atol=0.000018, equal_nan=True)

    assert z_df['weight'][0] is None
    assert z_df_eqw['weight'][0] == 'eqw'
    assert z_df_flux_gauss['weight'][0] == 'gauss_flux'
    assert z_df_strong['weight'][0] is None
    assert z_df_strong['lines'][0] == 'O3_5007A,H1_6563A'

    # Multi-index
    z_df = lime.redshift_calculation(obs.log)
    z_df_eqw = lime.redshift_calculation(obs.log, weight_parameter='eqw')
    z_df_flux_gauss = lime.redshift_calculation(obs.log, weight_parameter='gauss_flux')
    z_df_strong = lime.redshift_calculation(obs.log, line_list=['O3_5007A', 'H1_6563A'])

    assert np.allclose(z_df['z_mean'][0], 0.047526, atol=0.00024, equal_nan=True)
    assert np.allclose(z_df_eqw['z_mean'][0], 0.047526, atol=0.00024, equal_nan=True)
    assert np.allclose(z_df_flux_gauss['z_mean'][0], 0.047526, atol=0.00024, equal_nan=True)
    assert np.allclose(z_df_strong['z_mean'][0], 0.047498, atol=0.000018, equal_nan=True)

    assert np.all(z_df['weight'].to_numpy() == None)
    assert np.all(z_df_eqw['weight'] == 'eqw')
    assert np.all(z_df_flux_gauss['weight'] == 'gauss_flux')
    assert np.all(z_df_strong['weight'].to_numpy() == None)
    assert np.all(z_df_strong['lines'] == 'O3_5007A,H1_6563A')

    return


def test_unit_conversion():

    spec.unit_conversion(units_wave='nm', units_flux='Jy', norm_flux=1e-8)
    assert np.allclose(spec.wave.data, wave_array/10, equal_nan=True)
    assert np.allclose(spec.flux.data[:3], np.array([457.8036672 , 493.54866591, 493.17681153]), equal_nan=True)
    assert spec.norm_flux == 1e-8

    spec.unit_conversion(units_wave='A', units_flux='Flam', norm_flux=1)
    assert np.allclose(spec.wave.data, wave_array, equal_nan=True)
    assert np.allclose(spec.flux.data, flux_array, equal_nan=True)
    assert spec.norm_flux == 1

    return


def test_format_line_mask_option():

    array1 = format_line_mask_option('5000-5009', wave_array)
    array2 = format_line_mask_option('5000-5009,5876,6550-6570', wave_array)

    assert np.all(array1[0] == np.array([5000, 5009.]))

    assert np.all(array2[0] == np.array([5000, 5009.]))
    assert np.allclose(array2[1], np.array([5875.26214276, 5876.73785724]))
    assert np.all(array2[2] == np.array([6550, 6570.]))

    return


def test_refraction_index_air_vacuum():

    array1 = refraction_index_air_vacuum(wave_array)
    assert np.allclose(array1[:3], np.array([1.00030083, 1.00030082, 1.00030081]))

    return


import logging

import numpy as np
import pandas as pd
from numpy import array, abs, round, all, diff, char, searchsorted, unique, empty, arange, zeros
from .tools import DISPERSION_UNITS, UNITS_LATEX_DICT
from .io import _PARENT_BANDS, _LOG_EXPORT, _LOG_COLUMNS, check_file_dataframe, LiMe_Error
from pandas import DataFrame

_logger = logging.getLogger('LiMe')

_COMPs_KEYS = {'k': 'kinem',
               'p': 'profile_comp',
               't': 'transition_comp'}

ELEMENTS_DICT = dict(H='Hydrogen', He='Helium', Li='Lithium', Be='Beryllium', B='Boron', C='Carbon', N='Nitrogen', O='Oxygen', F='Fluorine',
                     Ne='Neon', Na='Sodium', Mg='Magnesium', Al='Aluminum', Si='Silicon', P='Phosphorus', S='Sulfur',
                     Cl='Chlorine', Ar='Argon', K='Potassium', Ca='Calcium', Sc='Scandium', Ti='Titanium', V='Vanadium',
                     Cr='Chromium', Mn='Manganese', Fe='Iron', Ni='Nickel', Co='Cobalt', Cu='Copper', Zn='Zinc',
                     Ga='Gallium', Ge='Germanium', As='Arsenic', Se='Selenium', Br='Bromine', Kr='Krypton', Rb='Rubidium',
                     Sr='Strontium', Y='Yttrium', Zr='Zirconium', Nb='Niobium', Mo='Molybdenum', Tc='Technetium',
                     Ru='Ruthenium', Rh='Rhodium', Pd='Palladium', Ag='Silver', Cd='Cadmium', In='Indium', Sn='Tin',
                     Sb='Antimony', Te='Tellurium', I='Iodine', Xe='Xenon', Cs='Cesium', Ba='Barium', La='Lanthanum',
                     Ce='Cerium', Pr='Praseodymium', Nd='Neodymium', Pm='Promethium', Sm='Samarium', Eu='Europium',
                     Gd='Gadolinium', Tb='Terbium', Dy='Dysprosium', Ho='Holmium', Er='Erbium', Tm='Thulium',
                     Yb='Ytterbium', Lu='Lutetium', Hf='Hafnium', Ta='Tantalum', W='Tungsten', Re='Rhenium', Os='Osmium',
                     Ir='Iridium', Pt='Platinum', Au='Gold', Hg='Mercury', Tl='Thallium', Pb='Lead', Bi='Bismuth',
                     Th='Thorium', Pa='Protactinium', U='Uranium', Np='Neptunium', Pu='Plutonium', Am='Americium',
                     Cm='Curium', Bk='Berkelium', Cf='Californium', Es='Einsteinium', Fm='Fermium', Md='Mendelevium',
                     No='Nobelium', Lr='Lawrencium', Rf='Rutherfordium', Db='Dubnium', Sg='Seaborgium', Bh='Bohrium',
                     Hs='Hassium', Mt='Meitnerium', Ds='Darmstadtium', Rg='Roentgenium', Cn='Copernicium', Nh='Nihonium',
                     Fl='Flerovium', Mc='Moscovium', Lv='Livermorium', Ts='Tennessine', Og='Oganesson')


VAL_LIST = [1000, 900, 500, 400, 100, 90, 50, 40, 10, 9, 5, 4, 1]
SYB_LIST = ["M", "CM", "D", "CD", "C", "XC", "L", "XL", "X", "IX", "V", "IV", "I"]


def int_to_roman(num):
    i, roman_num = 0, ''
    while num > 0:
        for _ in range(num // VAL_LIST[i]):
            roman_num += SYB_LIST[i]
            num -= VAL_LIST[i]
        i += 1
    return roman_num


def recover_ionization(string):

    part_len = len(string)
    index = part_len - 1  # Start at the last character
    while index >= 0 and string[index].isdigit():
        index -= 1

    # All characters
    if index == part_len - 1:
        particle, ionization = string, None

    # All numbers
    elif index < 0:
        particle, ionization = string, None

    # Particle and ionization
    else:
        particle, ionization = string[0:index+1], int(string[index+1:])

    return particle, ionization


def recover_transition(particle):

    # Check for PyNeb particle notation
    element = ELEMENTS_DICT.get(particle.symbol)
    if (element is not None) and (particle.ionization is not None):

        if particle.label in ('H1', 'He1', 'He2'):
            transition = 'rec'
        else:
            transition = 'col'
    else:
        transition = None

    return transition


def particle_notation(particle, transition):

    # Pre-assigned a transition time if particle and ionization
    if transition is None:
        transition = recover_transition(particle)

    if transition is not None:

        try:

            # Get ionization numeral
            ionization_roman = int_to_roman(particle.ionization)
            part_label = f'{particle.symbol}{ionization_roman}'

            # Collisional excited
            if transition == 'col':
                part_label = f'$[{part_label}]'

            # Semi-forbidden
            elif transition == 'sem':
                part_label = f'${part_label}]'

            # Recombination
            else:
                part_label = f'${part_label}'

        except:
            part_label = f'{particle}-$'

    else:
        part_label = f'{particle}-$'

    return part_label


def air_to_vacuum_function(input_array, sig_fig=None):

    input_array = np.array(input_array, ndmin=1)

    if 'U' in str(input_array.dtype): #TODO finde better way

        ion_array, wave_array, latex_array = label_decomposition(input_array)
        air_wave = wave_array
    else:
        air_wave = input_array

    refraction_index = (1 + 1e-6 * (287.6155 + 1.62887/np.power(air_wave*0.0001, 2) + 0.01360/np.power(air_wave*0.0001, 4)))
    output_array = (air_wave * 0.0001 * refraction_index) * 10000

    if sig_fig is not None:
        output_array = np.round(output_array, sig_fig) if sig_fig != 0 else np.round(output_array, sig_fig).astype(int)

    if 'U' in str(input_array.dtype):
        vacuum_wave = output_array.astype(str)
        output_array = np.core.defchararray.add(ion_array, '_')
        output_array = np.core.defchararray.add(output_array, vacuum_wave)
        output_array = np.core.defchararray.add(output_array, 'A')

    return output_array


def check_units_from_wave(str_wave):

    # One character unit
    units = str_wave[-1] if str_wave[-1] in DISPERSION_UNITS else None

    # Two characters unit
    if units is None:
        units = str_wave[-2:] if str_wave[-2:] in DISPERSION_UNITS else None

        # Warning for an unsuccessful notation
        if units is None:
            _logger.warning(f'The units from the transition "{str_wave}" could not be interpreted')
            wave = None
        else:
            wave = float(str_wave[:-2])

    else:
        wave = float(str_wave[:-1])

    return wave, units


def check_line_in_log(input_label, log=None, tol=1):

    # Guess the transition if only a wavelength provided
    if not isinstance(input_label, str):

        if log is not None:

            # Check the wavelength from the wave_obs column
            if ('wave_obs' in log.columns) and ('units_wave' in log.columns):
                ref_waves = log.wave_obs.values
                # units_wave = log.units_wave.values[0] # TODO add units to dataframe

            # Get it from the indexes
            else:
                ion_array, ref_waves = zip(*log.index.str.split('_').to_numpy())
                _wave0, units_wave = check_units_from_wave(ref_waves[0])
                ref_waves = char.strip(ref_waves, units_wave).astype(float)

            # Check if table rows are not sorted
            if not all(diff(ref_waves) >= 0): # TODO we might need to use something else than searchsorted
                _logger.warning(f'The lines log rows are not sorted from lower to higher wavelengths. This can cause '
                                f'issues to identify the lines using the transition wavelength. Try to use the string '
                                f'line label')

            # Locate the best candidate
            idx_closest = searchsorted(ref_waves, input_label)
            line_ion, line_wave = ion_array[idx_closest], ref_waves[idx_closest]

            # Check if input wavelength is close to the guessed line
            disc = abs(line_wave - input_label)
            if tol < disc < 1.5 * tol:
                _logger.info(f'The input line {input_label} has been identified with {log.iloc[idx_closest].name}')

            if disc > 1.5 * tol:
                _logger.warning(f'The closest line to the input line {input_label} is {log.iloc[idx_closest].name}. '
                                f'Please confirm that the database contains the target line and the units matched')

            input_label = log.iloc[idx_closest].name

        else:
            _logger.critical(f'The line "{input_label}" could not be identified: A lines log was not provided')

    return input_label


def latex_from_label(label, particle=None, wave=None, units_wave=None, kinem=None, transition_comp=None, scalar_output=False,
                     decimals=None):

    # Use input values if provided else compute from label
    if (particle is None) or (wave is None) or (wave is None) or (kinem is None) or (transition_comp is None):
        particle, wave, units_wave, kinem, profile_comp, transition_comp = label_composition(label)

    # Reshape to 1d array
    else:
        particle = array(particle, ndmin=1)
        wave = array(wave, ndmin=1)
        units_wave = array(units_wave, ndmin=1)
        kinem = array(kinem, ndmin=1)
        transition_comp = array(transition_comp, ndmin=1)

    n_items = wave.size
    latex_array = empty(n_items).astype('object')

    # Significant figures
    if decimals is None:
        wave = np.round(wave, 0).astype(int)
    else:
        wave = np.round(wave, decimals)

    for i in arange(n_items):

        # Particle label
        part_label = particle_notation(particle[i], transition_comp[i])

        # Wavelength and units label
        units_latex = UNITS_LATEX_DICT[units_wave[i]]

        # Kinematic label
        kinetic_comp = '' if kinem[i] == 0 else f'-k_{kinem[i]}'

        # Combine items
        latex_array[i] = f'{part_label}{wave[i]}{units_latex}{kinetic_comp}$'

    # Scalar output if requested and 1 length array
    if scalar_output:
        if latex_array.size == 1:
            latex_array = latex_array[0]

    return latex_array


def label_composition(line_list, ref_df=None, default_profile=None):

    # Empty containers for the label componentes
    n_comps = len(line_list)
    particle = [None] * n_comps #empty(n_comps).astype(str)
    wavelength = empty(n_comps)
    units_wave = empty(n_comps).astype(str)
    kinem = zeros(n_comps, dtype=int)
    profile_comp = [None] * n_comps
    transition_comp = [None] * n_comps

    # Loop through the components and get the components
    for i, line in enumerate(line_list):

        line_items = line.split('_')

        # Check the line has the items
        if len(line_items) < 2:
            raise LiMe_Error(f'The blend/merged component "{line}" in the transition list "{line_list}" does not have a'
                             f' recognised format. Please use a "Particle_WavelengthUnits" format in the configuration'
                             f'file')

        # Particle properties
        particle[i] = Particle.from_label(line_items[0])

        # Wavelength properties
        wavelength[i], units_wave[i] = check_units_from_wave(line_items[1])

        # Split the optional components: "H1_1216A_t-rec_k-0_p-g" -> {'t': 'rec', 'k': '0', 'p': 'g'}
        comp_conf = {optC[0]: optC[2:] for optC in line_items[2:]}

        # Kinematic component
        kinem[i] = int(comp_conf.get('k', 0))

        # Profile component
        profile_comp[i] = comp_conf.get('p', None)
        if profile_comp[i] is None:
            if default_profile is not None:
                profile_comp[i] = default_profile
            else:
                profile_comp[i] = 'g-emi'

        # Transition component
        trans = comp_conf.get('t', None)

        # If none is provided check from the table
        if (trans is None) and (ref_df is not None): # TODO check if try except is faster here
            if (line in ref_df.index) and ('transition' in ref_df.columns):
                trans = ref_df.loc[line, 'transition']

        # Else assume default
        if trans is None:
            trans = recover_transition(particle[i])

        transition_comp[i] = trans

    return particle, wavelength, units_wave, kinem, profile_comp, transition_comp


def label_decomposition(lines_list, bands=None, fit_conf=None, params_list=('particle', 'wavelength', 'latex_label'),
                        scalar_output=False):

    """
    This function takes a ``lines_list`` and returns several arrays with the requested parameters.

    If the user provides a `bands dataframe <https://lime-stable.readthedocs.io/en/latest/inputs/n_inputs3_line_bands.html>`_
    (``bands`` argument) dataframe and a `fitting documentation <https://lime-stable.readthedocs.io/en/latest/inputs/n_inputs4_fit_configuration.html>`_.
    (``fit_conf`` argument) the function will use this information to compute the requested outputs. Otherwise, only the
    `line label <https://lime-stable.readthedocs.io/en/latest/inputs/n_inputs2_line_labels.html>`_ will be used to derive
    the information.

    The ``params_list`` argument establishes the output parameters arrays. The options available are: "particle",
    "wavelength", "latex_label", "kinem", "profile_comp" and "transition_comp".

    If the ``lines_list`` argument only has one element the user can request an scalar output with ``scalar_output=True``.

    :param lines_list: Array of lines in LiMe notation.
    :type lines_list: list

    :param bands: Bands dataframe (or file address to the dataframe).
    :type bands: pandas.Dataframe, str, path.Pathlib, optional

    :param fit_conf: Fitting configuration.
    :type fit_conf: dict, optional

    :param params_list: List of output parameters. The default value is ('particle', 'wavelength', 'latex_label')
    :type params_list: tuple, optional

    :param scalar_output: Set to True for a Scalar output.
    :type scalar_output: bool

    """

    headers = ['particle', 'wavelength', 'latex_label', 'kinem', 'profile_comp', 'transition_comp']
    lines_df = pd.DataFrame(index=np.array(lines_list, ndmin=1), columns=headers)

    # Loop through the lines and derive their properties:
    for label in lines_df.index:
        line = Line(label, bands, fit_conf)

        lines_df.loc[label, :] = (line.particle[0].label, line.wavelength[0], line.latex_label[0], line.kinem[0],
                                  line.profile_comp[0], line.transition_comp[0])

    # Adjust column types
    lines_df['wavelength'] = pd.to_numeric(lines_df['wavelength'])

    # Recover the columns requested by the user
    output = []
    for i, param in enumerate(params_list):
        output.append(lines_df[param].to_numpy(copy=True))

    # If requested and single line, return the input as a scalar
    if scalar_output and (output[0].shape[0] == 1):
        output = tuple(item[0] for item in output)
    else:
        output = tuple(output)

    # TODO should we output as single array only if one property is requested?

    return output


class Particle:

    def __init__(self, label: str = None, symbol: str = None, ionization: int = str):

        self.label = label
        self.symbol = symbol
        self.ionization = ionization

        return

    @classmethod
    def from_label(cls, label):

        symbol, ionization = recover_ionization(label)

        return cls(label, symbol, ionization)

    def __str__(self):

        return self.label

    def __repr__(self):

        return self.label

    def __eq__(self, other):
        """Overrides the default implementation"""

        _equality = False
        if isinstance(other, Particle):
            if (other.label == self.label) and (other.symbol == self.symbol) and (other.ionization == self.ionization):
                _equality = True

        return _equality

    def __ne__(self, other):

        _inequality = True
        if isinstance(other, Particle):
            if (other.label == self.label) and (other.symbol == self.symbol) and (other.ionization == self.ionization):
                _inequality = False

        return _inequality


class Line:

    def __init__(self, label, band=None, fit_conf=None, profile=None, cont_from_bands=True, z_line=None,
                 interpret=True):

        # Label attributes
        self.label = label
        self.mask = None
        self.latex_label = None,
        self.profile_label, self.list_comps = np.nan, None

        self.particle = None
        self.wavelength, self.units_wave = None, None
        self.blended_check, self.merged_check = False, False

        self.kinem = None
        self.profile_comp = None
        self.transition_comp = None

        # Measurements attributes
        self.intg_flux, self.intg_flux_err = None, None
        self.peak_wave, self.peak_flux = None, None
        self.eqw, self.eqw_err = None, None
        self.gauss_flux, self.gauss_flux_err = None, None
        self.cont, self.std_cont =None, None
        self.m_cont, self.n_cont = None, None
        self.amp, self.center, self.sigma = None, None, None
        self.amp_err, self.center_err, self.sigma_err = None, None, None
        self.z_line = z_line
        self.v_r, self.v_r_err = None, None
        self.pixel_vel = None
        self.sigma_vel, self.sigma_vel_err = None, None
        self.sigma_thermal, self.sigma_instr = None, None
        self.snr_line, self.snr_cont = None, None
        self.observations, self.comments = 'no', 'no'
        self.pixel_mask = 'no'
        self.n_pixels = None
        self.FWHM_intg, self.FWHM_g, self.FWZI = None, None, None
        self.w_i, self.w_f = None, None
        self.v_med, self.v_50 = None, None
        self.v_5, self.v_10 = None, None
        self.v_90, self.v_95 = None, None
        self.v_1, self.v_99 = None, None
        self.chisqr, self.redchi = None, None
        self.aic, self.bic = None, None
        self.pixelWidth = None

        # Extra checks
        self._p_type = profile
        self._cont_from_adjacent = cont_from_bands
        self._decimal_wave = False
        self._narrow_check = False

        # Interpret the line from the user reference
        if interpret:
            self._from_label(label, band, fit_conf)

        return

    def _from_label(self, label, band=None, fit_conf=None):

        # If band is not provided use default database
        if band is None:
            band = _PARENT_BANDS
        band = check_file_dataframe(band, DataFrame, copy_input=False)

        # Check line in log
        self.label = check_line_in_log(label, band)

        # Copy the input configuration dictionary
        fit_conf = {} if fit_conf is None else fit_conf

        # Get label components
        comps_list = self.label.split('_')
        n_comps = len(comps_list)
        if n_comps < 2:
            raise LiMe_Error(f'The {self.label} the line label format is not recognized. '
                             f'Please use a "Particle_WavelengthUnits" format.')

        # Check the modularity      (only 2 comps)    (case b-6)
        modularity_comp = comps_list[-1] if (comps_list[-1] == 'b') or (comps_list[-1] == 'm') else None
        self._modularity_component(modularity_comp, fit_conf)

        # Review the components of the line
        ref_bands_df = band if isinstance(band, DataFrame) else None
        items = label_composition(self.list_comps, ref_df=ref_bands_df, default_profile=self._p_type)
        self.particle, self.wavelength, self.units_wave, self.kinem, self.profile_comp, self.transition_comp = items

        # Provide a bands from the log if possible
        if isinstance(band, DataFrame):

            # Query for the input label
            if self.label in band.index:
                self.mask = band.loc[self.label, 'w1':'w6'].to_numpy().astype(float)

            # Remove blended/merged suffix to check
            elif (self.blended_check or self.merged_check) and (self.label[:-2] in band.index):
                self.mask = band.loc[self.label[:-2], 'w1':'w6'].to_numpy().astype(float)

            # Could not find the mask
            else:
                self.mask = None

        else:
            self.mask = np.array(band)

        # Check if there are masked pixels in the line
        self.pixel_mask = fit_conf.get(f'{self.label}_mask', 'no')

        # Check if the wavelength has decimal transition
        self._decimal_wave = True if '.' in self.label else False

        # Compute latex entry if necessary
        self._review_latex_label(band)

        return

    @classmethod
    def from_log(cls, label, log=None, norm_flux=1):

        # Recover the label just in case
        label = check_line_in_log(label, log)

        # Create the line object
        inline = cls(label, interpret=False)

        if log is not None:

            if label in log.index:

                # Recover "simple" attributes
                for param in _LOG_EXPORT:

                    param_value = log.loc[label, param]

                    # Normalize
                    if _LOG_COLUMNS[param][0]:
                        param_value = param_value / norm_flux

                    inline.__setattr__(param, param_value)

                # Recover "complex" attributes
                for param in ['particle', 'wavelength', 'latex_label']:
                    inline.__setattr__(param, log.loc[label, param])

                # Band
                inline.mask = log.loc[label, 'w1':'w6'].values

                # Checks:
                if inline.profile_label is not np.nan:
                    inline.blended_check, inline.merged_check = False, False

                    # Merged line
                    if inline.label.endswith('_m'):
                        inline.merged_check = True
                        _, inline.units_wave = check_units_from_wave(label[:-1].split('_')[1])

                    # Blended line
                    else:
                        inline.blended_check = True
                        _, inline.units_wave = check_units_from_wave(label.split('_')[1])

                # Units transition
                else:
                    _, inline.units_wave = check_units_from_wave(label.split('_')[1])

                # List comps
                if inline.blended_check:
                    inline.list_comps = array(inline.profile_label.split('+'))

                else:
                    inline.list_comps = array([inline.label])

            else:
                _logger.warning(f'Input line {inline.label} not found in log')

        else:
            _logger.warning(f'No lines log introduced for the line {inline.label}')

        return inline

    def _modularity_component(self, modularity_label, fit_conf=None):

        # Not a single line
        if modularity_label is not None:

            if modularity_label == 'b':
                self.blended_check = True

            elif modularity_label == 'm':
                self.merged_check = True

            else:
                raise LiMe_Error(f'The modularity component {modularity_label} in the input line "{self.label}" is not'
                                 f' recognized')

        # Recover the profile components
        if self.merged_check or self.blended_check:

            # Check for profile label
            self.profile_label = np.nan if fit_conf is np.nan else fit_conf.get(self.label, np.nan)

            # Reset and warned the line has a suffix but there are no components provided
            if self.profile_label is np.nan:
                self.merged_check, self.blended_check = False, False

        # List of components only for blended
        if (self.merged_check or self.blended_check) and (self.profile_label is not np.nan):
            self.list_comps = self.profile_label.split('+') if self.blended_check else [self.label]

            # Check if there are repeated elements
            if len(self.list_comps) > 1:
                uniq, count = unique(self.list_comps, return_counts=True)
                if any(count > 1):
                    _logger.warning(f'There are repeated entries in the line label: {self.label} = {self.list_comps}')

        else:
            self.list_comps = [self.label]

        return

    def _review_latex_label(self, bands_df):

        # Check if there is a latex label on the database
        latex_exists = False
        if isinstance(bands_df, DataFrame):
            if 'latex_label' in bands_df.columns:
                if np.sum(bands_df.index.isin(self.list_comps)) == len(self.list_comps):
                    if not np.all(pd.isnull(bands_df.loc[self.list_comps, 'latex_label'])):
                        latex_exists = True

        # Merged
        if self.merged_check:
            if latex_exists:
                latex_list = list(bands_df.loc[self.list_comps, 'latex_label'].to_numpy())
            else:
                latex_list = list(latex_from_label(self.profile_label.split('+')))
            self.latex_label = array('+'.join(latex_list), ndmin=1)

        # Blended and single
        else:
            if latex_exists:
                self.latex_label = bands_df.loc[self.list_comps, 'latex_label'].to_numpy()
            else:
                self.latex_label = latex_from_label(None, self.particle, self.wavelength, self.units_wave, self.kinem,
                                                    self.transition_comp)

        return

    def __str__(self):

        return self.label

    def __repr__(self):

        return self.label



import numpy as np
from lime.model import gaussian_model
from matplotlib import pyplot as plt, rcParams
from matplotlib import font_manager
from pathlib import Path
import shutil
import matplotlib

matplotlib.font_manager._load_fontmanager(try_read_cache=False)

np.random.seed(2)

rcParams['font.family'] = ['MTF Saxy']

curve_dict = {'comp1': {'amp':0.75, 'center':1.00, 'sigma': 2.0},
              'comp2': {'amp':0.65, 'center':6.80, 'sigma': 1.8}}

cont = 0.6
err = 0.025
wave = np.linspace(-30, 30, 100)
wave_g = np.linspace(-30, 30, 1000)
noise = np.random.normal(0.0, err, size=wave.size)

flux_dict = {}
for curve, params in curve_dict.items():
        flux_dict[curve] = gaussian_model(wave, **params)
flux_comb = flux_dict['comp1'] + flux_dict['comp2'] + cont + noise

flux_dict_g = {}
for curve, params in curve_dict.items():
        flux_dict_g[curve] = gaussian_model(wave_g, **params)


fig, ax = plt.subplots(dpi=600)

w3, w4 = np.searchsorted(wave, (-3.0, 11))
w_cross1, w_cross2 = np.searchsorted(wave_g, (4.352, 3.942))

ax.step(wave[w3:w4], flux_comb[w3:w4], where='mid', color='black', linewidth=3)

# for curve, flux in flux_dict.items():
#         ax.plot(wave[w3:w4], flux[w3:w4] + cont, '--', linewidth=1)
ax.plot(wave_g[0:w_cross1], flux_dict_g['comp1'][0:w_cross1] + cont, '--', linewidth=1.5)
ax.plot(wave_g[w_cross2:-1], flux_dict_g['comp2'][w_cross2:-1] + cont, '--', linewidth=1.5)

dodge = 0 #- 0.1
residual = flux_comb - (flux_dict['comp1'] + flux_dict['comp2']) + dodge
ax.step(wave, residual, where='mid', color='black')

ax.fill_between(wave, -err+cont+dodge, err+cont+dodge, facecolor='tab:red', alpha=0.5)

# _ax.text(-1, 0.9, 'LiMe',
#         horizontalalignment='center',
#         verticalalignment='center',
#         fontsize=200, color='red', alpha=0.5)

ax.text(-13.5, 0.9, 'L',
        horizontalalignment='center',
        verticalalignment='center',
        fontsize=200, color='black')

ax.text(-5.5, 0.9, 'i',
        horizontalalignment='center',
        verticalalignment='center',
        fontsize=200, color='black')

ax.text(14, 0.9, 'e',
        horizontalalignment='center',
        verticalalignment='center',
        fontsize=200, color='black')

ax.set_ylim(0.4, 1.4)
ax.set_xlim(-18, 17)

ax.axis('off')
ax.axes.xaxis.set_visible(False)
ax.axes.yaxis.set_visible(False)

plt.tight_layout()
doc_images_folder = Path('../../docs/source/_static/')
plt.savefig(doc_images_folder/'logo_transparent.png', bbox_inches='tight', transparent=True)
plt.savefig(doc_images_folder/'logo_white.png', bbox_inches='tight')
print(f'Saving {doc_images_folder}')
# plt.show()



import numpy as np
import pandas as pd
import lime
from pathlib import Path

data_folder = Path(__file__).parent/'data_tests'
lines_log_address = data_folder/'manga_lines_log.txt'

# Data for the tests
lines_log = lime.load_log(lines_log_address)

class TestSampleClass:

    def test_from_sample_creation(self):

        sample1 = lime.Sample.from_file_list(id_list=['spec1', 'spec2'],
                                             log_list=[lines_log_address, lines_log_address],
                                             file_list=['spec1.fits', 'spec2.fits'])
        sample1.save_log(data_folder/f'sample1_3indeces.txt')

        assert list(sample1.log.index.names) == ['id', 'file', 'line']

        sample2 = lime.Sample.from_file_list(id_list=['spec1', 'spec2'], log_list=[lines_log_address, lines_log_address])
        sample2.save_log(data_folder/f'sample1_2indeces.txt')
        assert list(sample2.log.index.names) == ['id', 'line']

        sample3 = lime.Sample(data_folder/f'sample1_3indeces.txt')
        sample4 = lime.Sample(data_folder /f'sample1_2indeces.txt', levels=['id', 'line'])

        assert list(sample3.log.index.names) == ['id', 'file', 'line']
        assert list(sample4.log.index.names) == ['id', 'line']

        assert sample3.log.equals(sample1.log)
        assert sample4.log.equals(sample2.log)

        return

    # def test_multindex_log(self):
    #
    #     # Default import
    #     assert isinstance(obs.log.index, pd.MultiIndex)
    #     assert obs.log.index.names == ['id', 'line']
    #     assert np.all(obs.log.index.get_level_values('id').unique() == ['obj_0', 'obj_1', 'obj_2'])
    #
    #     return
    #
    # def test_extract_fluxes(self):
    #
    #     lime.extract_fluxes(obs.log, )
    #
    #     return


import numpy as np
import lime
from matplotlib import pyplot as plt
from pathlib import Path
import urllib.request
from astropy.wcs import WCS
from astropy.io import fits
import pytest

file_address = Path(__file__).parent/'data_tests'/'manga_spaxel.txt'
conf_file_address = Path(__file__).parent/'data_tests'/'manga.toml'
bands_file_address = Path(__file__).parent/'data_tests'/f'manga_line_bands.txt'
lines_log_address = Path(__file__).parent/'data_tests'/'manga_lines_log.txt'
cube_address = Path(__file__).parent/'data_tests'/'manga-8626-12704-LOGCUBE.fits.gz'
spatial_mask_address = Path(__file__).parent/'data_tests'/'SHOC579_mask.fits'
spatial_log_address = Path(__file__).parent/'data_tests'/'SHOC579_log.fits'

redshift = 0.0475
norm_flux = 1e-17
spaxel_label = '38-35'
cfg = lime.load_cfg(conf_file_address)

# Spaxel data
wave_array, flux_array, err_array = np.loadtxt(file_address, unpack=True)
pixel_mask = np.isnan(err_array)

spec = lime.Spectrum(wave_array, flux_array, err_array, redshift=redshift, norm_flux=norm_flux,
                     pixel_mask=pixel_mask)

spec.fit.frame(bands_file_address, cfg, id_conf_prefix=spaxel_label)

# MANGA cube web link and save file location
cube_url = 'https://data.sdss.org/sas/dr17/manga/spectro/redux/v3_1_1/8626/stack/manga-8626-12704-LOGCUBE.fits.gz'

# Download the cube file if not available (this may take some time)
if cube_address.is_file() is not True:
    urllib.request.urlretrieve(cube_url, cube_address)
    print(' Download completed!')
else:
    print('Observation found in folder')

# Open the MANGA cube fits file
with fits.open(cube_address) as hdul:
    wave = hdul['WAVE'].data
    flux_cube = hdul['FLUX'].data * norm_flux
    hdr = hdul['FLUX'].header

wcs = WCS(hdr)

# Define a LiMe cube object
cube = lime.Cube(wave, flux_cube, redshift=redshift, norm_flux=norm_flux, wcs=wcs)
# shoc579.plot.cube(6563, line_fg=4363)


class TestCubeClass:

    def test_read_cube(self):

        assert cube.norm_flux == norm_flux
        assert cube.redshift == redshift
        assert np.allclose(wave_array, cube.wave.data)
        assert np.allclose(wave_array / (1 + redshift), cube.wave_rest.data)

        return

    def test_spatial_masking(self):

        cube.spatial_masking('O3_4363A', param='SN_line', contour_pctls=[93, 96, 99],
                             output_address=spatial_mask_address)

        mask_dict = lime.load_spatial_mask(spatial_mask_address)

        assert len(mask_dict) == 3
        assert mask_dict['MASK_0'][0].sum() == mask_dict['MASK_0'][1]['NUMSPAXE']
        assert mask_dict['MASK_1'][0].sum() == mask_dict['MASK_1'][1]['NUMSPAXE']
        assert mask_dict['MASK_2'][0].sum() == mask_dict['MASK_2'][1]['NUMSPAXE']

        assert mask_dict['MASK_0'][1]['PARAM'] == 'SN_line'
        assert mask_dict['MASK_0'][1]['PARAMIDX'] == 99
        assert np.isclose(mask_dict['MASK_0'][1]['PARAMVAL'], 266.62467)

        assert mask_dict['MASK_1'][1]['PARAM'] == 'SN_line'
        assert mask_dict['MASK_1'][1]['PARAMIDX'] == 96
        assert np.isclose(mask_dict['MASK_1'][1]['PARAMVAL'], 49.29912)

        assert mask_dict['MASK_2'][1]['PARAM'] == 'SN_line'
        assert mask_dict['MASK_2'][1]['PARAMIDX'] == 93
        assert np.isclose(mask_dict['MASK_2'][1]['PARAMVAL'], 13.97339)

        return

    def test_get_spaxel(self):

        idx_j, idx_x = spaxel_label.split('-')
        spax = cube.get_spectrum(int(idx_j), int(idx_x))

        assert np.allclose(spax.wave, spec.wave.data)
        assert np.allclose(spax.flux, spec.flux.data)
        assert np.allclose(spax.redshift, spec.redshift)
        assert np.allclose(spax.norm_flux, spec.norm_flux)

        return

    def test_fit_spatial_mask(self):

        cfg['MASK_0_line_fitting']['bands'] = bands_file_address.as_posix()
        cube.fit.spatial_mask(spatial_mask_address, fit_conf=cfg, line_detection=True, mask_list=['MASK_0'],
                              output_address=spatial_log_address)

        assert spatial_log_address.is_file()

        spax_log = lime.load_log(spatial_log_address, page=f'{spaxel_label}_LINELOG')
        orig_log = lime.load_log(lines_log_address)

        # Test 3 lines # TODO review these fluxes
        assert np.sum(spax_log.index.isin(['O3_5007A', 'O3_5007A_k-1', 'He1_5016A'])) == 3
        assert np.isclose(spax_log.loc['O3_5007A', 'gauss_flux'], orig_log.loc['O3_5007A', 'gauss_flux'])
        assert np.isclose(spax_log.loc['O3_5007A_k-1', 'gauss_flux'], orig_log.loc['O3_5007A_k-1', 'gauss_flux'])
        assert np.isclose(spax_log.loc['He1_5016A', 'gauss_flux'], orig_log.loc['He1_5016A', 'gauss_flux'])

        return

    @pytest.mark.mpl_image_compare
    def test_plot_cube(self):

        fig = plt.figure()
        cube.plot.cube('H1_6563A', line_fg=4363, in_fig=fig)

        return fig

    @pytest.mark.mpl_image_compare
    def test_plot_cube_mask(self):

        fig = plt.figure()
        cube.plot.cube('H1_6563A', masks_file=spatial_mask_address, in_fig=fig)

        return fig

    @pytest.mark.mpl_image_compare
    def test_check_cube(self):

        fig = plt.figure()
        cube.check.cube('H1_6563A', lines_log_file=spatial_log_address, masks_file=spatial_mask_address, in_fig=fig)

        return fig

    def test_save_paramter_maps(self):

        ouput_folder = Path(__file__).parent/'data_tests'

        # Export the measurements log as maps:
        param_list = ['intg_flux', 'intg_flux_err', 'v_r', 'v_r_err']
        lines_list = ['H1_4861A', 'H1_6563A', 'O3_4363A', 'O3_4959A', 'O3_5007A']

        lime.save_parameter_maps(spatial_log_address, ouput_folder, param_list, lines_list,
                                 mask_file=spatial_mask_address, output_file_prefix='SHOC579_', wcs=wcs)

        param_list = ['gauss_flux', 'gauss_flux_err']
        lines_list = ['H1_4861A', 'H1_6563A', 'O3_4363A', 'O3_4959A', 'O3_5007A']
        lime.save_parameter_maps(spatial_log_address, ouput_folder, param_list, lines_list,
                                 mask_file=spatial_mask_address, wcs=wcs)

        intg_flux_file = ouput_folder/f'SHOC579_intg_flux.fits'
        gauss_flux_file = ouput_folder/f'gauss_flux.fits'

        assert intg_flux_file.is_file()
        assert gauss_flux_file.is_file()

        int_flux_map = fits.getdata(intg_flux_file, extname='O3_5007A')
        gauss_flux_map = fits.getdata(gauss_flux_file, extname='O3_5007A')

        assert np.isnan(int_flux_map).sum() == 5443
        assert np.isnan(gauss_flux_map).sum() == 5443

        orig_log = lime.load_log(lines_log_address)

        idx_j, idx_x = [int(item) for item in spaxel_label.split('-')]

        assert np.isclose(orig_log.loc['O3_5007A', 'intg_flux'], int_flux_map[idx_j, idx_x])
        assert np.isclose(orig_log.loc['O3_5007A', 'gauss_flux'], gauss_flux_map[idx_j, idx_x])

        return




import numpy as np
import pandas as pd
import lime

def test_line_bands():

    log0 = lime.line_bands()
    parent_bands = lime.load_log(lime._lines_database_path)

    # TODO rework on the master database
    # assert np.all(log0.index == parent_bands.index)
    assert np.all(log0.columns == parent_bands.columns)
    # assert log0.equals(parent_bands)

    log1 = lime.line_bands(wave_intvl=(3000, 7000))
    assert np.all((3000 <= log1.wavelength.to_numpy()) & (log1.wavelength.to_numpy() <= 7000))

    log2 = lime.line_bands(wave_intvl=(3000, 7000), z_intvl=(0, 2))
    assert np.all((3000 * (1 + 0) <= log2.wavelength.to_numpy()) & (log2.wavelength.to_numpy() <= 7000 * (1 + 2)))

    log3 = lime.line_bands(lines_list=['O3_4363A', 'O3_4959A', 'O3_5007A'])
    assert log3.index.isin(['O3_4363A', 'O3_4959A', 'O3_5007A']).sum() == log3.index.size

    log4 = lime.line_bands(particle_list=['O3', 'S2'])
    assert log4.particle.isin(['O3', 'S2']).sum() == log4.index.size

    return


__all__ = [
    'load_fits',
    'save_cfg',
    'load_cfg',
    'load_log',
    'load_spatial_mask',
    'save_log',
    'log_parameters_calculation',
    'log_to_HDU',
    'save_parameter_maps',
    '_LOG_EXPORT_RECARR',
    '_LOG_EXPORT',
    '_LOG_COLUMNS']

import os
import configparser
import logging
import numpy as np
import pandas as pd

from . import Error, __version__
from sys import exit, stdout, version_info
from pathlib import Path
from distutils.util import strtobool
from collections.abc import Sequence

from astropy.io import fits
from astropy.table import Table

from .tables import table_fluxes


try:
    import openpyxl
    openpyxl_check = True
    from openpyxl.utils.dataframe import dataframe_to_rows

except ImportError:
    openpyxl_check = False

try:
    import asdf
    asdf_check = True
except ImportError:
    asdf_check = False


try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib

try:
    import toml
    toml_check = True
except ImportError:
    toml_check = False

_logger = logging.getLogger('LiMe')

# Reading file with the format and export status for the measurements
_params_table_file = Path(__file__).parent/'resources/types_params.txt'
_PARAMS_CONF_TABLE = pd.read_csv(_params_table_file, delim_whitespace=True, header=0, index_col=0)

_LINES_DATABASE_FILE = Path(__file__).parent/'resources/parent_bands.txt'

# Dictionary with the parameter formart
_LOG_COLUMNS = dict(zip(_PARAMS_CONF_TABLE.index.values,
                        _PARAMS_CONF_TABLE.loc[:, 'Norm_by_flux':'dtype'].values))

# Parameters notation latex formatDictionary with the parameter formart
_LOG_COLUMNS_LATEX = dict(zip(_PARAMS_CONF_TABLE.index.values,
                          _PARAMS_CONF_TABLE.loc[:, 'latex_label'].values))

# Array with the parameters to be included in the output log
_LOG_EXPORT = _PARAMS_CONF_TABLE.loc[_PARAMS_CONF_TABLE.Export_log.to_numpy().astype(bool)].index.to_numpy()
_LOG_EXPORT_TYPES = _PARAMS_CONF_TABLE.loc[_PARAMS_CONF_TABLE.Export_log.to_numpy().astype(bool)].dtype.to_numpy()
_LOG_EXPORT_DICT = dict(zip(_LOG_EXPORT, _LOG_EXPORT_TYPES))
_LOG_EXPORT_RECARR = np.dtype(list(_LOG_EXPORT_DICT.items()))

# Attributes with measurements for log
_ATTRIBUTES_FIT = _PARAMS_CONF_TABLE.loc[_PARAMS_CONF_TABLE.Fit_attributes.to_numpy().astype(bool)].index.to_numpy()
_RANGE_ATTRIBUTES_FIT = np.arange(_ATTRIBUTES_FIT.size)

# Dictionary with the parameter dtypes
_LOG_TYPES_DICT = dict(zip(_PARAMS_CONF_TABLE.index.to_numpy(),
                           _PARAMS_CONF_TABLE.dtype.to_numpy()))

# Numpy recarray dtype for the pandas dataframe creation

GLOBAL_LOCAL_GROUPS = ['line_fitting', 'chemical_model'] # TODO not implemented

FLUX_TEX_TABLE_HEADERS = [r'$Transition$', '$EW(\AA)$', '$F(\lambda)$', '$I(\lambda)$']
FLUX_TXT_TABLE_HEADERS = [r'$Transition$', 'EW', 'EW_error', 'F(lambda)', 'F(lambda)_error', 'I(lambda)', 'I(lambda)_error']

KIN_TEX_TABLE_HEADERS = [r'$Transition$', r'$Comp$', r'$v_{r}\left(\nicefrac{km}{s}\right)$', r'$\sigma_{int}\left(\nicefrac{km}{s}\right)$', r'Flux $(\nicefrac{erg}{cm^{-2} s^{-1} \AA^{-1})}$']
KIN_TXT_TABLE_HEADERS = [r'$Transition$', r'$Comp$', 'v_r', 'v_r_error', 'sigma_int', 'sigma_int_error', 'flux', 'flux_error']

# TODO replace this error with the one of .ini
class LiMe_Error(Exception):
    """LiMe exception function"""


# Function to load SpecSyzer configuration file
def load_cfg(file_address, fit_cfg_suffix='_line_fitting'):

    """

    This function reads a configuration file with the `toml format <https://toml.io/en/>`_. The text file extension
    must adhere to this format specifications to be successfully read.

    If one of the file sections has the suffix specified by the ``fit_cfg_suffix`` this function will query its items and
    convert the entries to the format expected by `LiMe functions <https://lime-stable.readthedocs.io/en/latest/inputs/n_inputs4_fit_configuration.html>`_.
    The default suffix is "_line_fitting".

    The function will show a critical warning if it fails to convert an item in a ``fit_cfg_suffix`` section.

    :param file_address: Input configuration file address.
    :type file_address: str, pathlib.Path

    :param fit_cfg_suffix: Suffix for LiMe configuration sections. The default value is "_line_fitting".
    :type fit_cfg_suffix:  str

    :return: Parsed configuration data
    :type: dict

    """

    file_path = Path(file_address)

    # Open the file
    if file_path.is_file():

        # Toml file
        with open(file_path, mode="rb") as fp:
            cfg_lime = tomllib.load(fp)

    else:
        raise LiMe_Error(f'The configuration file was not found at: {file_address}')

    # Convert the configuration entries from the string format if possible
    if fit_cfg_suffix is not None:
        for section, items in cfg_lime.items():
            if section.endswith(fit_cfg_suffix):
                for i_key, i_value in items.items():
                    try:
                        cfg_lime[section][i_key] = format_option_value(i_value, i_key, section)
                    except:
                        _logger.critical(f'Failure to convert entry: "{i_key} = {i_value}" at section [{section}] ')

    return cfg_lime


# Function to save SpecSyzer configuration file
def save_cfg(param_dict, output_file, section_name=None, clear_section=False):

    """
    This function safes the input dictionary into a configuration file. If no section is provided the input dictionary
    overwrites the data

    """
    # TODO for line_fitting/model save dictionaries as line
    output_path = Path(output_file)

    if output_path.suffix == '.toml':
        # TODO review convert numpy arrays and floats64
        if toml_check:
            toml_str = toml.dumps(param_dict)
            with open(output_path, "w") as f:
                f.write(toml_str)
        else:
            raise LiMe_Error(f'toml library is not installed. Toml files cannot be saved')

    # Creating a new file (overwritting old if existing)
    else:

        if section_name == None:

            # Check all entries are dictionaries
            values_list = [*param_dict.values()]
            section_check = all(isinstance(x, dict) for x in values_list)
            assert section_check, f'ERROR: Dictionary for {output_file} cannot be converted to configuration file. Confirm all its values are dictionaries'

            output_cfg = configparser.ConfigParser()
            output_cfg.optionxform = str

            # Loop throught he sections and options to create the files
            for section_name, options_dict in param_dict.items():
                output_cfg.add_section(section_name)
                for option_name, option_value in options_dict.items():
                    option_formatted = formatStringOutput(option_value, option_name, section_name)
                    output_cfg.set(section_name, option_name, option_formatted)

            # Save to a text format
            with open(output_file, 'w') as f:
                output_cfg.write(f)

        # Updating old file
        else:

            # Confirm file exists
            file_check = os.path.isfile(output_file)

            # Load original cfg
            if file_check:
                output_cfg = configparser.ConfigParser()
                output_cfg.optionxform = str
                output_cfg.read(output_file)

            # Create empty cfg
            else:
                output_cfg = configparser.ConfigParser()
                output_cfg.optionxform = str

            # Clear section upon request
            if clear_section:
                if output_cfg.has_section(section_name):
                    output_cfg.remove_section(section_name)

            # Add new section if it is not there
            if not output_cfg.has_section(section_name):
                output_cfg.add_section(section_name)

            # Map key values to the expected format and store them
            for option_name, option_value in param_dict.items():
                option_formatted = formatStringOutput(option_value, option_name, section_name)
                output_cfg.set(section_name, option_name, option_formatted)

            # Save to a text file
            with open(output_file, 'w') as f:
                output_cfg.write(f)

    return


def load_log(file_address, page: str = 'LINELOG', levels: list = ['id', 'line']):

    """
    This function reads the input ``file_address`` as a pandas dataframe.

    The expected file types are ".txt", ".pdf", ".fits", ".asdf" and ".xlsx". The dataframes expected format is discussed
    on the `line bands <https://lime-stable.readthedocs.io/en/latest/inputs/n_inputs3_line_bands.html>`_ and `measurements <https://lime-stable.readthedocs.io/en/latest/inputs/n_inputs4_fit_configuration.html>`_ documentation.

    For ".fits" and ".xlsx" files the user can provide a page name ``ext`` for the HDU/sheet. The default name is "_LINELOG".

    To reconstruct a `MultiIndex dataframe <https://pandas.pydata.org/docs/user_guide/advanced.html#advanced-hierarchical>`_
    the user needs to specify the ``sample_levels``.

    :param file_address: Input log address.
    :type file_address: str, Path

    :param page: Name of the HDU/sheet for ".fits"/".xlsx" files. The default value is "_LINELOG".
    :type page: str, optional

    :param levels: Indexes name list for MultiIndex dataframes. The default value is ['id', 'line'].
    :type levels: list, optional

    :return: lines log table
    :rtype: pandas.DataFrame

    """

    # Check file is at path
    log_path = Path(file_address)
    if not log_path.is_file():
        raise LiMe_Error(f'No lines log found at {file_address}\n')

    file_name, file_type = log_path.name, log_path.suffix

    try:

        # Fits file:
        if file_type == '.fits':
            log = Table.read(log_path, page, character_as_bytes=False).to_pandas()
            log.set_index('index', inplace=True)

            # Change 'nan' to np.nan
            idcs_nan_str = log['profile_label'] == 'nan'
            log.loc[idcs_nan_str, 'profile_label'] = np.nan

        # Excel table
        elif file_type in ['.xlsx' or '.xls']:
            log = pd.read_excel(log_path, sheet_name=page, header=0, index_col=0)

        # ASDF file
        elif file_type == '.asdf':
            with asdf.open(log_path) as af:
                log_RA = af[page]
                log = pd.DataFrame.from_records(log_RA, columns=log_RA.dtype.names)
                log.set_index('index', inplace=True)

                # Change 'nan' to np.nan
                idcs_nan_str = log['profile_label'] == 'nan'
                log.loc[idcs_nan_str, 'profile_label'] = np.nan

        # Text file
        elif file_type == '.txt':
            log = pd.read_csv(log_path, delim_whitespace=True, header=0, index_col=0, comment='#')

        elif file_type == '.csv':
            log = pd.read_csv(log_path, sep=';', delim_whitespace=False, header=0, index_col=0)

        else:
            _logger.warning(f'File type {file_type} is not recognized. This can cause issues reading the log.')
            log = pd.read_csv(log_path, delim_whitespace=True, header=0, index_col=0)

    except ValueError as e:
        exit(f'\nERROR: LiMe could not open {file_type} file at {log_path}\n{e}')

    # Restore levels if multi-index
    if log.columns.isin(levels).sum() == len(levels):
        log.set_index(levels, inplace=True)

    return log


def save_log(dataframe, file_address, page='LINELOG', parameters='all', header=None, store_version=True):

    """

    This function saves the input ``log_dataframe`` at the ``file_address`` provided by the user.

    The accepted extensions are ".txt", ".pdf", ".fits", ".asdf" and ".xlsx".

    For ".fits" and ".xlsx" files the user can provide a page name for the HDU/sheet with the ``ext`` argument.
    The default name is "LINELOG".

    The user can specify the ``parameters`` to be saved in the output file.

    For ".fits" files the user can provide a dictionary to add to the ``fits_header``.

    :param dataframe: Lines log dataframe.
    :type dataframe: pandas.DataFrame

    :param file_address: Output log address.
    :type file_address: str, Path

    :param parameters: Output parameters list. The default value is "all"
    :type parameters: list

    :param page: Name of the HDU/sheet for ".fits"/".xlsx" files.
    :type page: str, optional

    :param header: Dictionary for ".fits" and ".asdf" file headers.
    :type header: dict, optional

    :param store_version: Save LiMe version as footnote or page header on the output log. The default value is True.
    :type store_version: bool, optional

    """

    # Confirm file path exits
    log_path = Path(file_address)
    assert log_path.parent.exists(), f'- ERROR: Output lines log folder not found: {log_path.parent}'
    file_name, file_type = log_path.name, log_path.suffix

    if len(dataframe.index) > 0:

        # In case of multi-index dataframe
        if isinstance(dataframe.index, pd.MultiIndex):
            log = dataframe.reset_index()
        else:
            log = dataframe

        # Slice the log if the user provides a list of columns
        if parameters != 'all':
            parameters_list = np.array(parameters, ndmin=1)
            lines_log = log[parameters_list]
            param_dtypes = [_LOG_EXPORT_RECARR[param] for param in parameters_list]

        else:
            lines_log = log
            param_dtypes = list(_LOG_TYPES_DICT.values())

        # Default txt log with the complete information
        if file_type == '.txt':
            with open(log_path, 'wb') as output_file:
                pd.set_option('multi_sparse', False)
                string_DF = lines_log.to_string()
                string_DF = string_DF if store_version is False else string_DF + f'\n#LiMe_{__version__}'
                output_file.write(string_DF.encode('UTF-8'))

        elif file_type == '.csv':
            with open(log_path, 'wb') as output_file:
                pd.set_option('multi_sparse', False)
                string_DF = lines_log.to_csv(sep=';', na_rep='NaN')
                output_file.write(string_DF.encode('UTF-8'))

        # Pdf fluxes table # TODO error while saving all parameters
        elif file_type == '.pdf':
            table_fluxes(lines_log, log_path.parent/log_path.stem, header_format_latex=_LOG_COLUMNS_LATEX,
                         lines_notation=log.latex_label.values, store_version=store_version)

        # Log in a fits format
        elif file_type == '.fits':
            if isinstance(lines_log, pd.DataFrame):

                if store_version:
                    header = {} if header is None else header
                    header['LiMe'] = __version__

                lineLogHDU = log_to_HDU(lines_log, ext_name=page, header_dict=header)

                if log_path.is_file(): # TODO this strategy is slow for many inputs
                    try:
                        fits.update(log_path, data=lineLogHDU.data, header=lineLogHDU.header, extname=lineLogHDU.name, verify=True)
                    except KeyError:
                        fits.append(log_path, data=lineLogHDU.data, header=lineLogHDU.header, extname=lineLogHDU.name)
                else:
                    hdul = fits.HDUList([fits.PrimaryHDU(), lineLogHDU])
                    hdul.writeto(log_path, overwrite=True, output_verify='fix')

        # Log in excel format
        elif file_type == '.xlsx' or file_type == '.xls':

            # Check openpyxl is installed else leave
            if openpyxl_check:

                # New excel
                if not log_path.is_file():

                    with pd.ExcelWriter(log_path) as writer:
                        lines_log.to_excel(writer, sheet_name=page)

                        if store_version:
                            df_empty = pd.DataFrame()
                            df_empty.to_excel(writer, sheet_name=f'LiMe_{__version__}', index=False)

                # Updating existing file
                else:
                    wb = openpyxl.load_workbook(log_path)

                    # Remove if existing and create anew
                    if page in wb.sheetnames:
                        wb.remove(wb[page])
                    sheet = wb.create_sheet(page, index=0)

                    # Add data one row at a time
                    for i, row in enumerate(dataframe_to_rows(lines_log, index=True, header=True)):
                        if len(row) > 1: # TOFIX Remove the frozen list logic
                            sheet.append(row)

                    # Save the data
                    wb.save(log_path)

                    # Add new sheet

                    # df_old = load_log(log_path)
                    # # with pd.ExcelWriter(log_path, engine='openpyxl') as writer:
                    # #
                    # #     book = openpyxl.load_workbook(log_path)
                    # #     writer.book = book
                    # #     writer.sheets = dict((ws.title, ws) for ws in book.worksheets)
                    # #     lines_log.to_excel(writer, sheet_name=page, index=True)
                    #     # dataframe.to_excel(writer, sheet_name=page, index=True)
                    #
                    #     # if store_version:
                    #     #     df_empty = pd.DataFrame()
                    #     #     df_empty.to_excel(writer, sheet_name=f'LiMe_{__version__}')
                    #
                    #     # book.close()

            else:
                _logger.critical(f'openpyxl is not installed. Lines log {file_address} could not be saved')

        # Advance Scientific Storage Format
        elif file_type == '.asdf':

            tree = {page: lines_log.to_records(index=True, column_dtypes=_LOG_TYPES_DICT, index_dtypes='<U50')}

            # Create new file
            if not log_path.is_file():
                af = asdf.AsdfFile(tree)
                af.write_to(log_path)

            # Update file
            else:
                with asdf.open(log_path, mode='rw') as af:
                    af.tree.update(tree)
                    af.update()

        # Extension not recognised
        else:
            raise LiMe_Error(f'output extension "{file_type}" was not recognised for file {log_path}')

    # Empty log
    else:
        _logger.info('The output log has no measurements. An output file will not be saved')

    return


def results_to_log(line, log, norm_flux):

    # Loop through the line components
    for i, comp in enumerate(line.list_comps):

        # Add bands wavelengths
        log.at[comp, 'w1'] = line.mask[0]
        log.at[comp, 'w2'] = line.mask[1]
        log.at[comp, 'w3'] = line.mask[2]
        log.at[comp, 'w4'] = line.mask[3]
        log.at[comp, 'w5'] = line.mask[4]
        log.at[comp, 'w6'] = line.mask[5]

        # Treat every line
        for j in _RANGE_ATTRIBUTES_FIT:

            param = _ATTRIBUTES_FIT[j]

            # Get component parameter
            param_value = line.__getattribute__(param)
            if _LOG_COLUMNS[param][2] and (param_value is not None):
                param_value = param_value[i]

            # De-normalize
            if _LOG_COLUMNS[param][0]:
                if param_value is not None:
                    param_value = param_value * norm_flux

            # Just string for particle
            if j == 7:
                param_value = param_value.label

            # Converting None entries to str (9 = profile_label)
            if j == 9:
                if param_value is None:
                    param_value = 'None'

            log.at[comp, param] = param_value

    return


def check_file_dataframe(df_variable, variable_type, ext='LINELOG', sample_levels=('id', 'line'), copy_input=True):

    if isinstance(df_variable, variable_type):
        if copy_input:
            output = df_variable.copy()
        else:
            output = df_variable

    elif isinstance(df_variable, (str, Path)):
        input_path = Path(df_variable)
        if input_path.is_file():
            output = load_log(df_variable, page=ext, levels=sample_levels)
        else:
            output = None

    else:
        output = df_variable

    return output


def check_file_configuration(input_cfg, default_cfg_section=None, local_cfg_section=None, cfg_suffix='_model_fitting'):

    if isinstance(input_cfg, dict):
        cfg_dict = input_cfg.copy()
    else:
        cfg_dict = load_cfg(input_cfg, fit_cfg_suffix=None)

    # Get default configuration and updated with local if necessary
    if default_cfg_section is not None:
        if f'{default_cfg_section}{cfg_suffix}' in cfg_dict:
            output = {**cfg_dict[f'{default_cfg_section}{cfg_suffix}']}

            if local_cfg_section is not None:
                if f'{local_cfg_section}{cfg_suffix}' in cfg_dict:
                    output = {**output, **cfg_dict[f'{local_cfg_section}{cfg_suffix}']}

    else:
        output = cfg_dict

    return output


_parent_bands_file = Path(__file__).parent/'resources/parent_bands.txt'
_PARENT_BANDS = load_log(_parent_bands_file)

# Function to check if variable can be converte to float else leave as string
def check_numeric_Value(s):
    try:
        output = float(s)
        return output
    except ValueError:
        return s


# Function to map a string to its variable-type
def format_option_value(entry_value, key_label, section_label='', float_format=None, nan_format='nan'):

    output_variable = entry_value

    # # None variable
    # if (entry_value == 'None') or (entry_value is None):
    #     output_variable = None

    # Dictionary blended lines
    if isinstance(entry_value, str):

        output_variable = {}

        try:
            keys_and_values = entry_value.split(',')
            for pair in keys_and_values:

                # Conversion for parameter class atributes
                if ':' in pair:
                    key, value = pair.split(':')
                    if value == 'None':
                        output_variable[key] = None
                    elif key in ['value', 'min', 'max']:
                        output_variable[key] = float(value)
                    elif key == 'vary':
                        output_variable[key] = strtobool(value) == 1
                    else:
                        output_variable[key] = value
                elif '_mask' in key_label:
                    output_variable = entry_value
                # Conversion for non-parameter class atributes (only str conversion possible)
                else:
                    output_variable = check_numeric_Value(entry_value)
        except:
            raise LiMe_Error(f'Failure to convert configuration entry: {key_label} = {entry_value} in section {section_label}')

    return output_variable


def log_parameters_calculation(input_log, parameter_list, formulae_list):

    # Load the log if necessary file:
    file_check = False
    if isinstance(input_log, pd.DataFrame):
        log_df = input_log

    elif isinstance(input_log, (str, Path)):
        file_check = True
        log_df = load_log(input_log)

    else:
        _logger.critical(
            f'Not a recognize log format. Please use a pandas dataframe or a string/Path object for file {input_log}')
        exit(1)

    # Parse the combined expression
    expr = ''
    for col, formula in zip(parameter_list, formulae_list):
        expr += f'{col}={formula}\n'

    # Compute the new parameters
    log_df.eval(expr=expr, inplace=True)

    # Save to the previous location
    if file_check:
        save_log(log_df, input_log)


    return

def extract_wcs_header(wcs, drop_axis=None):

    if wcs is not None:

        # Remove 3rd dimensional axis if present
        if drop_axis is not None:
            if drop_axis == 'spectral':
                input_wcs = wcs.dropaxis(2) if wcs.naxis == 3 else wcs
            elif drop_axis == 'spatial':
                if wcs.naxis == 3:
                    input_wcs = wcs.dropaxis(1)
                    input_wcs = wcs.dropaxis(0)
            else:
                raise LiMe_Error(f'Fits coordinates axis: "{drop_axis}" not recognized. Please use: "spectral" or'
                                 f' "spatial"')

        else:
            input_wcs = wcs

        # Convert to HDU header
        hdr_coords = input_wcs.to_fits()[0].header

    else:
        hdr_coords = None

    return hdr_coords


def log_to_HDU(log, ext_name=None, column_types={}, header_dict={}):

    # For non empty logs
    if not log.empty:

        if len(column_types) == 0:
            params_dtype = _LOG_TYPES_DICT
        else:
            params_dtype = _LOG_TYPES_DICT.copy()
            user_dtype = column_types.copy()
            params_dtype.update(user_dtype)

        linesSA = log.to_records(index=True, column_dtypes=params_dtype, index_dtypes='<U50')
        linesCol = fits.ColDefs(linesSA)

        hdr = fits.Header(header_dict) if isinstance(header_dict, dict) else header_dict
        linesHDU = fits.BinTableHDU.from_columns(linesCol, name=ext_name, header=hdr)

    # Empty log
    else:
        # TODO create and empty HDU
        linesHDU = None

    return linesHDU


def load_fits(file_address, instrument, frame_idx=None):

    if instrument == 'ISIS':

        if frame_idx is None:
            frame_idx = 0

        # Open fits file
        with fits.open(file_address) as hdul:
            data, header = hdul[frame_idx].data, hdul[frame_idx].header

        assert 'ISIS' in header['INSTRUME'], 'Input spectrum instrument '

        # William Herschel Telescope ISIS instrument
        w_min = header['CRVAL1']
        dw = header['CD1_1']  # dw = 0.862936 INDEF (Wavelength interval per pixel)
        pixels = header['NAXIS1']  # nw = 3801 number of output pixels
        w_max = w_min + dw * pixels
        wave = np.linspace(w_min, w_max, pixels, endpoint=False)

        return wave, data, header

    elif instrument == 'fits-cube':

        # Open fits file
        with fits.open(file_address) as hdul:
            data, hdr = hdul[frame_idx].data, hdul[frame_idx].header


        dw = hdr['CD3_3']
        w_min = hdr['CRVAL3']
        nPixels = hdr['NAXIS3']
        w_max = w_min + dw * nPixels
        wave = np.linspace(w_min, w_max, nPixels, endpoint=False)

        print('CD3_3', dw)
        print('CRVAL3', w_min)
        print('NAXIS3', nPixels)
        print('np.diff', np.diff(wave).mean(), np.std(np.diff(wave)))

        return wave, data, hdr

    elif instrument == 'OSIRIS':

        # Default _frame index
        if frame_idx is None:
            frame_idx = 0

        # Open fits file
        with fits.open(file_address) as hdul:
            data, header = hdul[frame_idx].data, hdul[frame_idx].header

        # assert 'OSIRIS' in header['INSTRUME']

        w_min = header['CRVAL1']
        dw = header['CD1_1']  # dw (Wavelength interval per pixel)
        pixels = header['NAXIS1']  # nw number of output pixels
        w_max = w_min + dw * pixels
        wave = np.linspace(w_min, w_max, pixels, endpoint=False)

        return wave, data, header

    elif instrument == 'SDSS':

        # Open fits file
        with fits.open(file_address) as hdul:
            data, header_0, header_2, header_3 = hdul[1].data, hdul[0].header, hdul[2].data, hdul[3].data

        assert 'SDSS 2.5-M' in header_0['TELESCOP']

        wave = 10.0 ** data['loglam']
        SDSS_z = float(header_2["z"][0] + 1)
        wave_rest = wave / SDSS_z

        flux_norm = data['flux']
        flux = flux_norm / 1e17

        headers = (header_0, header_2, header_3)

        # return wavelength_array, flux, headers
        return wave, data, headers

    elif instrument == 'xshooter':

        # Default _frame index
        if frame_idx is None:
            frame_idx = 1

        # Following the steps at: https://archive.eso.org/cms/eso-data/help/1dspectra.html
        with fits.open(file_address) as hdul:
            data, header = hdul[frame_idx].data, hdul[frame_idx].header

        w_min = header['CRVAL1']
        dw = header['CDELT1']  # dw (Wavelength interval per pixel)
        pixels = header['NAXIS1']  # nw number of output pixels
        w_max = w_min + dw * pixels
        wave = np.linspace(w_min, w_max, pixels, endpoint=False)

        # wave = data[0][0]

        return wave, data, header

    elif instrument == 'MEGARA':

        # Default _frame index
        if frame_idx is None:
            frame_idx = 1

        # Following the steps at: https://archive.eso.org/cms/eso-data/help/1dspectra.html
        with fits.open(file_address) as hdul:
            data, header = hdul[frame_idx].data, hdul[frame_idx].header

        w_min = header['CRVAL1']
        dw = header['CDELT1']  # dw (Wavelength interval per pixel)
        pixels = header['NAXIS1']  # nw number of output pixels
        w_max = w_min + dw * pixels
        wave = np.linspace(w_min, w_max, pixels, endpoint=False)

        return wave, data, header

    else:

        print('-- WARNING: Instrument not recognize')

        # Open fits file
        with fits.open(file_address) as hdul:
            data, header = hdul[frame_idx].data, hdul[frame_idx].header

        return data, header


def formatStringOutput(value, key, section_label=None, float_format=None, nan_format='nan'):

    # TODO this one should be the default option
    # TODO add more cases for dicts
    # Check None entry
    if value is not None:

        # Check string entry
        if isinstance(value, str):
            formatted_value = value

        else:

            # Case of an array
            scalarVariable = True
            if isinstance(value, (Sequence, np.ndarray)):

                # Confirm is not a single value array
                if len(value) == 1:
                    value = value[0]

                # Case of an array
                else:
                    scalarVariable = False
                    formatted_value = ','.join([str(item) for item in value])

            if scalarVariable:

                # Case single float
                if isinstance(value, str):
                    formatted_value = value
                else:
                    if np.isnan(value):
                        formatted_value = nan_format
                    else:
                        formatted_value = str(value)

    else:
        formatted_value = 'None'

    return formatted_value


def progress_bar(i, i_max, post_text, n_bar=10):

    # Size of progress bar
    j = i/i_max
    stdout.write('\r')
    message = f"[{'=' * int(n_bar * j):{n_bar}s}] {int(100 * j)}% {post_text}"
    stdout.write(message)
    stdout.flush()

    return

def save_parameter_maps(lines_log_file, output_folder, param_list, line_list, mask_file=None, mask_list='all',
                        image_shape=None, log_ext_suffix='_LINELOG', spaxel_fill_value=np.nan, output_file_prefix=None,
                        header=None, wcs=None):

    """

    This function converts a line measurements log from an IFS cube, into a set of 2D parameter maps.

    The parameter ".fits" files are saved into the ``output_folder``. These files are named after the parameters in the
    ``param_list`` with the optional prefix from the ``output_file_prefix`` argument. These files will have one page per
    line in the ``line_list`` argument.

    The user can provide a spatial mask file address from which to recover the spaxels with line measurements. If the
    mask ``.fits`` file contains several pages, the user can provide a ``mask_list`` with the ones to explore. Otherwise,
    all mask pages will be used.

    .. attention::
        The user can provide an ``image_shape`` tuple to generate the output parameter map. However, for a large image
        size this approach may require a long time to query the log file pages.

    The expected page name in the input ``lines_log_file`` is "idx_j-idx_i_log_ext_suffix" where "idx_j" and "idx_i"
    are the y and x array coordinates of the cube coordinates, by default ``log_ext_suffix='_LINELOG'``.

    The output ``.fits`` parameter page header includes the ``PARAM`` and ``LINE`` entries with the line and parameter
    labels respectively. The user should also include a ``wcs`` argument to export the astronomical coordinates to the
    output files. The user can add additional information via the ``header`` argument.

    :param lines_log_file: Fits file with IFU cube line measurements.
    :param lines_log_file: str, pathlib.Path

    :param param_list: List of parameters to map
    :param param_list: list

    :param line_list: List of lines to map
    :param line_list: list

    :param output_folder: Output folder to save the maps
    :param output_folder: str, pathlib.Path

    :param mask_file: Address of binary spatial mask file
    :type mask_file: str, pathlib.Path

    :param mask_list: Mask name list to explore on the ``mask_file``.
    :type mask_list: list, optional

    :param image_shape: Array with the image spatial size.
    :param image_shape: list, tuple, optional

    :param spaxel_fill_value: Map filling value for empty pixels. The default value is "numpy.nan".
    :param spaxel_fill_value: float, optional

    :param log_ext_suffix: Suffix for the lines log extension. The default value is "_LINELOG"
    :param log_ext_suffix: str, optional

    :param output_file_prefix: Prefix for the output parameter ".fits" file. The default value is None.
    :param output_file_prefix: str, optional

    :param header: Dictionary for parameter ".fits" file header
    :type header: dict, optional

    :param wcs: Observation `world coordinate system <https://docs.astropy.org/en/stable/wcs/index.html>`_.
    :type wcs: astropy WCS, optional

    """

    assert Path(lines_log_file).is_file(), f'- ERROR: lines log at {lines_log_file} not found'
    assert Path(output_folder).is_dir(), f'- ERROR: Output parameter maps folder {output_folder} not found'

    # Compile the list of voxels to recover the provided masks
    if mask_file is not None:

        assert Path(mask_file).is_file(), f'- ERROR: mask file at {mask_file} not found'

        with fits.open(mask_file) as maskHDUs:

            # Get the list of mask extensions
            if mask_list == 'all':
                if ('PRIMARY' in maskHDUs) and (len(maskHDUs) > 1):
                    mask_list = []
                    for i, HDU in enumerate(maskHDUs):
                        mask_name = HDU.name
                        if mask_name != 'PRIMARY':
                            mask_list.append(mask_name)
                    mask_list = np.array(mask_list)
                else:
                    mask_list = np.array(['PRIMARY'])
            else:
                mask_list = np.array(mask_list, ndmin=1)

            # Combine all the mask voxels into one
            for i, mask_name in enumerate(mask_list):
                if i == 0:
                    mask_array = maskHDUs[mask_name].data
                    image_shape = mask_array.shape
                else:
                    assert image_shape == maskHDUs[mask_name].data.shape, '- ERROR: Input masks do not have the same dimensions'
                    mask_array += maskHDUs[mask_name].data

            # Convert to boolean
            mask_array = mask_array.astype(bool)

            # List of spaxels in list [(idx_j, idx_i), ...] format
            spaxel_list = np.argwhere(mask_array)

    # No mask file is provided and the user just defines an image size tupple (nY, nX)
    else:
        mask_array = np.ones(image_shape).astype(bool)
        spaxel_list = np.argwhere(mask_array)

    # Generate containers for the data:
    images_dict = {}
    for param in param_list:

        # Make sure is an array and loop throuh them
        for line in line_list:
            images_dict[f'{param}-{line}'] = np.full(image_shape, spaxel_fill_value)

    # Loop through the spaxels and fill the parameter images
    n_spaxels = spaxel_list.shape[0]
    spaxel_range = np.arange(n_spaxels)

    with fits.open(lines_log_file) as logHDUs:

        for i_spaxel in spaxel_range:
            idx_j, idx_i = spaxel_list[i_spaxel]
            spaxel_ref = f'{idx_j}-{idx_i}{log_ext_suffix}'

            progress_bar(i_spaxel, n_spaxels, post_text=f'of spaxels from file ({lines_log_file}) read ({n_spaxels} total spaxels)')

            # Confirm log extension exists
            if spaxel_ref in logHDUs:

                # Recover extension data
                log_data = logHDUs[spaxel_ref].data
                log_lines = log_data['index']

                # Loop through the parameters and the lines:
                for param in param_list:
                    idcs_log = np.argwhere(np.in1d(log_lines, line_list))
                    for i_line in idcs_log:
                        images_dict[f'{param}-{log_lines[i_line][0]}'][idx_j, idx_i] = log_data[param][i_line][0]

    # New line after the rustic progress bar
    print()

    # Recover coordinates from the wcs to store in the headers:
    hdr_coords = extract_wcs_header(wcs, drop_axis='spectral')

    # Save the parameter maps as individual fits files with one line per page
    output_file_prefix = '' if output_file_prefix is None else output_file_prefix
    for param in param_list:

        # Primary header
        paramHDUs = fits.HDUList()
        paramHDUs.append(fits.PrimaryHDU())

        # ImageHDU for the parameter maps
        for line in line_list:

            # Create page header with the default data
            hdr_i = fits.Header()
            hdr_i['LINE'] = (line, 'Line label')
            hdr_i['PARAM'] = (param, 'LiMe parameter label')

            # Add WCS information
            if hdr_coords is not None:
                hdr_i.update(hdr_coords)

            # Add user information
            if header is not None:
                page_hdr = header.get(f'{param}-{line}', None)
                page_hdr = header if page_hdr is None else page_hdr
                hdr_i.update(page_hdr)

            # Create page HDU entry
            HDU_i = fits.ImageHDU(name=line, data=images_dict[f'{param}-{line}'], header=hdr_i, ver=1)
            paramHDUs.append(HDU_i)

        # Write to new file
        output_file = Path(output_folder)/f'{output_file_prefix}{param}.fits'
        paramHDUs.writeto(output_file, overwrite=True, output_verify='fix')

    return


def load_spatial_mask(mask_file, mask_list=None, return_coords=False):

    # Masks array container
    spatial_mask_dict = {}

    # Limit the analysis to some masks
    mask_list_check = False if mask_list is None else True

    # Check that mask is there:
    if mask_file is not None:

        mask_file = Path(mask_file)

        if mask_file.is_file():

            with fits.open(mask_file) as maskHDULs:

                # Save the fits data to restore later
                counter = 0
                for HDU in maskHDULs:
                    ext_name = HDU.name
                    if HDU.name != 'PRIMARY':
                        if mask_list_check:
                            if ext_name in mask_list:
                                spatial_mask_dict[ext_name] = (HDU.data.astype('bool'), HDU.header)
                                counter += 1
                        else:
                            spatial_mask_dict[ext_name] = (HDU.data.astype('bool'), HDU.header)
                            counter += 1

                # Warn if the mask file does not contain any of the expected masks
                if counter == 0:
                    _logger.warning(f'No masks extensions were found in file {mask_file}')

        else:
            _logger.warning(f'No mask file was found at {mask_file.as_posix()}.')

    # Return the mask as a set of coordinates (no headers)
    if return_coords:
        for mask_name, mask_data in spatial_mask_dict.items():
            idcs_spaxels = np.argwhere(mask_data[0])
            spatial_mask_dict[mask_name] = idcs_spaxels

    return spatial_mask_dict


def check_file_array_mask(var, mask_list=None):

    # Check if file
    if isinstance(var, (str, Path)):

        input = Path(var)
        if input.is_file():
            mask_dict = load_spatial_mask(var, mask_list)
        else:
            raise Error(f'No spatial mask file at {Path(var).as_posix()}')

    # Array
    elif isinstance(var, (np.ndarray, list)):

        # Re-adjust the variable
        var = np.ndarray(var, ndmin=3)
        masks_array = np.squeeze(np.array_split(var, var.shape[0], axis=0))

        # Confirm boolean array
        if masks_array.dtype != bool:
            _logger.warning(f'The input mask array should have a boolean variables (True/False)')

        # Incase user gives a str
        mask_list = [mask_list] if isinstance(mask_list, str) else mask_list

        # Check if there is a mask list
        if len(mask_list) == 0:
            mask_list = [f'SPMASK{i}' for i in range(masks_array.shape[0])]

        # Case the number of masks names and arrays is different
        elif masks_array.shape[0] != len(mask_list):
            _logger.warning(f'The number of input spatial mask arrays is different than the number of mask names')

        # Everything is fine
        else:
            mask_list = mask_list

        # Create mask dict with empty headers
        mask_dict = dict(zip(mask_list, (masks_array, {})))

    else:

        raise Error(f'Input mask format {type(input)} is not recognized for a mask file. Please declare a fits file, a'
                    f' numpy array or a list/array of numpy arrays')

    return mask_dict



import numpy as np
import pandas as pd
import logging

from pathlib import Path
from scipy import signal
from sys import exit
from lmfit.models import PolynomialModel
from inspect import signature
from .io import LiMe_Error, check_file_dataframe
from .transitions import label_decomposition
import astropy.units as au

try:
    import joblib
    joblib_check = True
except ImportError:
    joblib_check = False


_logger = logging.getLogger('LiMe')

MACHINE_PATH = Path(__file__).parent/'resources'/'LogitistRegression_v2_cost1_logNorm.joblib'

# TODO replace this mechanic
WAVE_UNITS_DEFAULT, FLUX_UNITS_DEFAULT = au.AA, au.erg / au.s / au.cm ** 2 / au.AA


try:
    from specutils import Spectrum1D, SpectralRegion
    from specutils.manipulation import noise_region_uncertainty
    from specutils.fitting import find_lines_derivative
    specutils_check = True

except ImportError:
    specutils_check = False


def compute_line_width(idx_peak, spec_flux, delta_i, min_delta=2, emission_check=True):
    """
    Algororithm to measure emision line width given its peak location
    :param idx_peak:
    :param spec_flux:
    :param delta_i:
    :param min_delta:
    :return:
    """

    i = idx_peak

    if emission_check:
        while (spec_flux[i] > spec_flux[i + delta_i]) or (np.abs(idx_peak - (i + delta_i)) <= min_delta):
            i += delta_i
    else:
        while (spec_flux[i] < spec_flux[i + delta_i]) or (np.abs(idx_peak - (i + delta_i)) <= min_delta):
            i += delta_i

    return i


class LineFinder:

    def __init__(self, machine_model_path=MACHINE_PATH):

        # self.ml_model = joblib.load(machine_model_path) # THIS CAN be warning at opening the file

        return

    def continuum_fitting(self, degree_list=[3, 7, 7, 7], threshold_list=[5, 3, 2, 2], plot_results=False,
                          return_std=False):

        # Check for a masked array
        if np.ma.is_masked(self.flux):
            mask_cont = ~self.flux.mask
            input_wave, input_flux = self.wave.data, self.flux.data
        else:
            mask_cont = np.ones(self.flux.size).astype(bool)
            input_wave, input_flux = self.wave, self.flux

        # Loop through the fitting degree
        for i, degree in enumerate(degree_list):

            # Establishing the flux limits
            input_flux_selection = input_flux[mask_cont]
            if np.ma.isMaskedArray(input_flux_selection): # Bugged numpy
                low_lim, high_lim = np.nanpercentile(input_flux_selection.filled(np.nan), (16, 84))
            else:
                low_lim, high_lim = np.percentile(input_flux_selection, (16, 84))

            low_lim, high_lim = low_lim / threshold_list[i], high_lim * threshold_list[i]

            # Add new entries to the mask
            mask_cont = mask_cont & (input_flux >= low_lim) & (input_flux <= high_lim)

            poly3Mod = PolynomialModel(prefix=f'poly_{degree}', degree=degree)
            poly3Params = poly3Mod.guess(input_flux[mask_cont], x=input_wave[mask_cont])

            try:
                poly3Out = poly3Mod.fit(input_flux[mask_cont], poly3Params, x=input_wave[mask_cont])
                continuum_fit = poly3Out.eval(x=input_wave)

            except TypeError:
                _logger.warning(f'- The continuum fitting polynomial as degree ({degree}) is larger than data points'
                                f' number')
                continuum_fit = np.full(input_wave.size, np.nan)

            # Compute the continuum and assign replace the value outside the bands the new continuum
            if plot_results:
                title = f'Continuum fitting, iteration ({i+1}/{len(degree_list)})'
                continuum_full = poly3Out.eval(x=self.wave.data)
                self.plot._plot_continuum_fit(continuum_full, mask_cont, low_lim, high_lim, threshold_list[i], title)

        # Include the standard deviation of the spectrum for the unmasked pixels
        if return_std:
            std_spec = np.std((self.flux-continuum_fit)[mask_cont])
            output_params = (continuum_fit, std_spec)
        else:
            output_params = continuum_fit

        return output_params

    def peak_detection(self, limit_threshold=None, continuum=None, distance=4, ml_mask=None, plot_results=False):

        # No user imput provided compute the intensity threshold from the 84th percentil
        limit_threshold = np.percentile(self.flux, 84) if limit_threshold is None else limit_threshold
        limit_threshold = limit_threshold + continuum if continuum is not None else limit_threshold

        # Index the intensity peaks
        # input_flux = self.flux.data if np.ma.is_masked()
        # mask_valid = ~self.flux.mask if np.ma.is_masked(self.flux) else np.ones(self.flux.data.size).astype(bool)
        # peak_fp, _ = signal.find_peaks(self.flux.data[mask_valid], height=limit_threshold[mask_valid], distance=distance)

        peak_fp, _ = signal.find_peaks(self.flux, height=limit_threshold, distance=distance)

        # Plot the results
        if plot_results:
            self.plot._plot_peak_detection(peak_fp, limit_threshold, continuum, ml_mask=ml_mask,
                                      plot_title='Peak detection results ')

        return peak_fp

    def ml_line_detection(self, continuum, box_width=11):

        # Normalize the flux
        input_flux = self.flux if not np.ma.is_masked(self.flux) else self.flux.data
        input_flux = np.log10((input_flux/continuum - 1) + 10)

        # Reshape to the training dimensions
        input_flux = np.array(input_flux, ndmin=2)

        # Container for the true pixels
        detection_mask = np.zeros(self.flux.size).astype(bool)

        # Case of 1D
        spectrum_pixels = input_flux.shape[1]
        for i in np.arange(spectrum_pixels):
            if i + box_width <= spectrum_pixels:
                y = input_flux[:, i:i + box_width]
                if not np.any(np.isnan(y)):
                    detection_mask[i:i + box_width] = detection_mask[i:i + box_width] | self.ml_model.predict(y)[0]
                    # print(f'y {i} ({np.sum(y)}): {self.ml_model.predict(y)[0]}')

        return detection_mask

    def line_detection(self, bands=None, cont_fit_degree=(3, 7, 7, 7), cont_int_thres=(5, 3, 2, 2), noise_sigma_factor=3,
                       line_type='emission', width_tol=5, band_modification=None, ml_detection=None, plot_cont_calc=False,
                       plot_peak_calc=False):

        """

        This function compares the input lines bands in the observation spectrum to confirm the presence of lines.

        The input bands can be specified as a pandas dataframe or the path to its file via the ``bands_df`` argument.

        Prior to the line detection, the spectrum continuum is fit in an iterative process. The ``cont_fit_degree`` array
        provides the order of the fitting polynomial, while the ``cont_int_thres`` array provides the threshold intensity
        factor for the threshold flux above and below the continuum to exclude at each iteration.

        After the continuum has been normalized, the ``noise_sigma_factor`` establishes the standard deviation factor
        beyond which a positive line detection is assumed.

        The additional arguments provide additional tools to adjust the line detection and show the steps/results.

        :param bands: Input bands dataframe or the address to its file.
        :type bands: pandas.Dataframe, str, pathlib.Path

        :param cont_fit_degree: Continuum polynomial fitting degree.
        :type cont_fit_degree: tuple, optional

        :param cont_int_thres: Continuum maximum intensity threshold to include pixels.
        :type cont_int_thres: tuple, optional

        :param noise_sigma_factor: Continuum standard deviation factor for line detection. The default value is 3.
        :type noise_sigma_factor: float, optional

        :param line_type: Line type. The default value is "emission".
        :type line_type: str, optional

        :param width_tol: Minimum number of pixels between peaks/troughs. The default value is 5.
        :type width_tol: float, optional

        :param band_modification: Method to adjust the line band with. The default value is None.
        :type band_modification: str, optional

        :param ml_detection: Machine learning algorithm to detect lines. The default value is None.
        :type ml_detection: str, optional

        :param plot_cont_calc: Plot the continuum fit at each iteration. The default value is False
        :type plot_cont_calc: bool, optional

        :param plot_peak_calc: Plot the detected peaks/troughs. The default value is False
        :type plot_peak_calc: bool, optional

        """

        # TODO input log should not be None... Maybe read database

        # Fit the continuum
        cont_flux, cond_Std = self.continuum_fitting(degree_list=cont_fit_degree, threshold_list=cont_int_thres,
                                                     return_std=True, plot_results=plot_cont_calc)

        # Check via machine learning algorithm
        if ml_detection is not None:
            if joblib_check:
                self.ml_model = joblib.load(MACHINE_PATH)
                ml_mask = self.ml_line_detection(cont_flux) if ml_detection else None
            else:
                raise ImportError(f'Need to install joblib library to use machine learning detection')
        else:
            ml_mask = None

        # Check for the peaks of the emission lines
        detec_min = noise_sigma_factor * cond_Std
        idcs_peaks = self.peak_detection(detec_min, cont_flux, plot_results=plot_peak_calc, ml_mask=ml_mask)

        # Compare against the theoretical values
        bands = check_file_dataframe(bands, pd.DataFrame)
        if bands is not None:

            # Match peaks with theoretical lines
            matched_DF = self.label_peaks(idcs_peaks, bands, width_tol=width_tol, band_modification=band_modification,
                                          line_type=line_type)

            return matched_DF

        else:

            return

    def label_peaks(self, peak_table, mask_df, line_type='emission', width_tol=5, band_modification=None, detect_check=False):

        # TODO auto param should be changed to boolean
        # Establish the type of input values for the peak indexes, first numpy array
        if isinstance(peak_table, np.ndarray):
            idcsLinePeak = peak_table

        # Specutils table
        else:
            # Query the lines from the astropy finder tables #
            if len(peak_table) != 0:
                idcsLineType = peak_table['line_type'] == line_type
                idcsLinePeak = np.array(peak_table[idcsLineType]['line_center_index'])
            else:
                idcsLinePeak = np.array([])

        # Security check in case no lines detected
        if len(idcsLinePeak) == 0:
            return pd.DataFrame(columns=mask_df.columns)

        # Prepare dataframe to stored the matched lines
        matched_DF = pd.DataFrame.copy(mask_df)
        matched_DF['signal_peak'] = np.nan

        # Theoretical wave values
        waveTheory = label_decomposition(matched_DF.index.values, params_list=['wavelength'])[0]
        matched_DF['wavelength'] = waveTheory

        # Match the lines with the theoretical emission
        tolerance = np.diff(self.wave_rest).mean() * width_tol
        matched_DF['observation'] = 'not_detected'
        unidentifiedLine = dict.fromkeys(matched_DF.columns.values, np.nan)

        # Get the wavelength peaks
        wave_peaks = self.wave_rest[idcsLinePeak]

        # Only treat pixels outisde the masks
        if np.ma.is_masked(wave_peaks):
            idcsLinePeak = idcsLinePeak[~wave_peaks.mask]
            wave_peaks = wave_peaks[~wave_peaks.mask].data

        for i in np.arange(wave_peaks.size):

            idx_array = np.where(np.isclose(a=waveTheory.astype(np.float64), b=wave_peaks[i], atol=tolerance))

            if len(idx_array[0]) == 0:
                unknownLineLabel = 'xy_{:.0f}A'.format(np.round(wave_peaks[i]))

                # Scheme to avoid repeated lines
                if (unknownLineLabel not in matched_DF.index) and detect_check:
                    newRow = unidentifiedLine.copy()
                    newRow.update({'wavelength': wave_peaks[i], 'w3': wave_peaks[i] - 5, 'w4': wave_peaks[i] + 5,
                                   'observation': 'not_identified'})
                    matched_DF.loc[unknownLineLabel] = newRow

            else:

                row_index = matched_DF.index[matched_DF.wavelength == waveTheory[idx_array[0][0]]]
                matched_DF.loc[row_index, 'observation'] = 'detected'
                matched_DF.loc[row_index, 'signal_peak'] = idcsLinePeak[i]
                theoLineLabel = row_index[0]

                blended_check = True if '_b' in theoLineLabel else False
                minSeparation = 4 if blended_check else 2

                # Width is only computed for blended lines
                if band_modification is not None:
                    if band_modification == 'auto':
                        if blended_check is False:
                            emission_check = True if line_type == 'emission' else False
                            idx_min = compute_line_width(idcsLinePeak[i], self.flux, delta_i=-1, min_delta=minSeparation, emission_check=emission_check)
                            idx_max = compute_line_width(idcsLinePeak[i], self.flux, delta_i=1, min_delta=minSeparation, emission_check=emission_check)
                            matched_DF.loc[row_index, 'w3'] = self.wave_rest[idx_min]
                            matched_DF.loc[row_index, 'w4'] = self.wave_rest[idx_max]

        # Include_only_detected
        idcs_unknown = matched_DF['observation'] == 'not_detected'
        matched_DF.drop(index=matched_DF.loc[idcs_unknown].index.values, inplace=True)

        # Sort by wavelength
        matched_DF.sort_values('wavelength', inplace=True)
        matched_DF.drop(columns=['wavelength', 'observation'], inplace=True)

        return matched_DF


# Line finder default parameters
LINE_DETECT_PARAMS = signature(LineFinder.line_detection).parameters
LINE_DETECT_PARAMS = {key: value.default for key, value in LINE_DETECT_PARAMS.items()}
LINE_DETECT_PARAMS.pop('self')



import numpy as np
import lime
from pathlib import Path
import pytest
from matplotlib import pyplot as plt
from matplotlib.testing.compare import compare_images
from lime.io import _LOG_EXPORT_DICT
from os import remove

# Data for the tests
baseline_folder = Path(__file__).parent/'data_tests'
file_address = Path(__file__).parent/'data_tests'/'manga_spaxel.txt'
conf_file_address = Path(__file__).parent/'data_tests'/'manga.toml'
bands_file_address = Path(__file__).parent/'data_tests'/f'manga_line_bands.txt'
lines_log_address = Path(__file__).parent/'data_tests'/'manga_lines_log.txt'

redshift = 0.0475
norm_flux = 1e-17
cfg = lime.load_cfg(conf_file_address)

wave_array, flux_array, err_array = np.loadtxt(file_address, unpack=True)
pixel_mask = np.isnan(err_array)

spec = lime.Spectrum(wave_array, flux_array, err_array, redshift=redshift, norm_flux=norm_flux,
                     pixel_mask=pixel_mask)

spec.fit.frame(bands_file_address, cfg, id_conf_prefix='38-35')


class TestSpectrumClass:

    def test_read_spectrum(self):

        assert spec.norm_flux == norm_flux
        assert spec.redshift == redshift
        assert np.allclose(wave_array, spec.wave.data)
        assert np.allclose(wave_array / (1 + redshift), spec.wave_rest.data)
        assert np.allclose(flux_array, spec.flux.data * norm_flux, equal_nan=True)
        assert np.allclose(err_array, spec.err_flux.data * norm_flux, equal_nan=True)

        return

    def test_change_spectrum(self):

        # Cropping

        spec0 = lime.Spectrum(wave_array, flux_array, err_array, redshift=redshift, norm_flux=norm_flux,
                             pixel_mask=pixel_mask)
        spec1 = lime.Spectrum(wave_array, flux_array, err_array, redshift=redshift, norm_flux=norm_flux,
                             pixel_mask=pixel_mask, crop_waves=(5500, -1))
        spec2 = lime.Spectrum(wave_array, flux_array, err_array, redshift=redshift, norm_flux=norm_flux,
                             pixel_mask=pixel_mask, crop_waves=(0, 9000))
        spec3 = lime.Spectrum(wave_array, flux_array, err_array, redshift=redshift, norm_flux=norm_flux,
                             pixel_mask=pixel_mask, crop_waves=(5500, 9000))

        assert spec0.wave.data[0] == wave_array[0]
        assert spec0.wave.data[-1] == wave_array[-1]

        assert np.isclose(spec1.wave.data[0], 5500, atol=np.diff(spec1.wave.data).mean())
        assert spec1.wave.data[-1] == wave_array[-1]

        assert spec2.wave.data[0] == wave_array[0]
        assert np.isclose(spec2.wave.data[-1], 9000, atol=np.diff(spec2.wave.data).mean())

        assert np.isclose(spec3.wave.data[0], 5500, atol=np.diff(spec3.wave.data).mean())
        assert np.isclose(spec3.wave.data[-1], 9000, atol=np.diff(spec3.wave.data).mean())

        assert spec0.redshift == redshift

        # Updating redshift
        spec0.update_redshift(redshift=0.06)
        assert spec0.redshift == 0.06
        assert spec0.norm_flux == norm_flux
        assert np.allclose(wave_array, spec0.wave.data)
        assert np.allclose(wave_array / (1 + 0.06), spec0.wave_rest.data)
        assert np.allclose(flux_array, spec0.flux.data * norm_flux, equal_nan=True)
        assert np.allclose(err_array, spec0.err_flux.data * norm_flux, equal_nan=True)

        spec0.update_redshift(redshift=redshift)
        assert spec0.norm_flux == norm_flux
        assert np.allclose(wave_array, spec0.wave.data)
        assert np.allclose(wave_array / (1 + redshift), spec0.wave_rest.data)
        assert np.allclose(flux_array, spec0.flux.data * norm_flux, equal_nan=True)
        assert np.allclose(err_array, spec0.err_flux.data * norm_flux, equal_nan=True)

        return

    @pytest.mark.line_detection_plot
    def test_line_detection_plot(self):

        match_bands = spec.line_detection(bands_file_address, cont_fit_degree=[3, 7, 7, 7], cont_int_thres=[5, 3, 2, 1.5])

        fig = plt.figure()
        spec.plot.spectrum(in_fig=fig, line_bands=match_bands)

        return fig

    @pytest.mark.mpl_image_compare
    def test_plot_spectrum(self):

        fig = plt.figure()
        spec.plot.spectrum(in_fig=fig)

        return fig

    @pytest.mark.mpl_image_compare
    def test_plot_spectrum_with_fits(self):

        fig = plt.figure()
        spec.plot.spectrum(in_fig=fig, include_fits=True)

        return fig

    @pytest.mark.mpl_image_compare
    def test_check_bands_spectrum(self):

        fig = plt.figure()
        spec.check.bands(bands_file=bands_file_address, in_fig=fig)

        return fig

    @pytest.mark.mpl_image_compare
    def test_plot_spectrum_maximize(self):

        fig = plt.figure()
        spec.plot.spectrum(in_fig=fig, include_fits=True, maximize=True)

        return fig

    @pytest.mark.mpl_image_compare
    def test_plot_spectrum_with_bands(self):

        fig = plt.figure()
        spec.plot.spectrum(in_fig=fig, line_bands=bands_file_address)

        return fig

    @pytest.mark.mpl_image_compare
    def test_plot_line(self):

        fig = plt.figure()
        spec.plot.bands('Fe3_4658A_p-g_emis', in_fig=fig)

        return fig

    @pytest.mark.mpl_image_compare
    def test_plot_grid(self):

        fig = plt.figure()
        spec.plot.grid(in_fig=fig)

        return fig

    @pytest.mark.mpl_image_compare(tolerance=10)
    def test_plot_cinematics(self):

        fig = plt.figure()
        spec.plot.velocity_profile('O3_5007A', in_fig=fig)

        return fig

    def test_measurements_txt_file(self):

        extension = 'txt'
        spec.save_log(baseline_folder/f'test_lines_log.{extension}')

        log_orig = lime.load_log(lines_log_address)
        log_test = lime.load_log(baseline_folder/f'test_lines_log.{extension}')

        for line in spec.log.index:
            for param in spec.log.columns:

                # String
                if _LOG_EXPORT_DICT[param].startswith('<U'):
                    if log_orig.loc[line, param] is np.nan:
                        assert log_orig.loc[line, param] is log_test.loc[line, param]
                    else:
                        assert log_orig.loc[line, param] == log_test.loc[line, param]

                # Float
                else:
                    if param not in ['eqw', 'eqw_err']:
                        assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.05, equal_nan=True)
                    else:
                        assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.15,
                                              equal_nan=True)

        return

    def test_measurements_fits_file(self):

        extension = 'fits'
        spec.save_log(baseline_folder/f'test_lines_log.{extension}')

        log_orig = lime.load_log(lines_log_address)
        log_test = lime.load_log(baseline_folder/f'test_lines_log.{extension}')

        for line in spec.log.index:
            for param in spec.log.columns:

                # String
                if _LOG_EXPORT_DICT[param].startswith('<U'):
                    if log_orig.loc[line, param] is np.nan:
                        assert log_orig.loc[line, param] is log_test.loc[line, param]
                    else:
                        assert log_orig.loc[line, param] == log_test.loc[line, param]

                # Float
                else:
                    if param not in ['eqw', 'eqw_err']:
                        assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.05,
                                              equal_nan=True)
                    else:
                        assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.15,
                                              equal_nan=True)

        return

    def test_measurements_csv_file(self):

        extension = 'csv'
        spec.save_log(baseline_folder/f'test_lines_log.{extension}')

        log_orig = lime.load_log(lines_log_address)
        log_test = lime.load_log(baseline_folder/f'test_lines_log.{extension}')

        for line in spec.log.index:
            for param in spec.log.columns:

                # String
                if _LOG_EXPORT_DICT[param].startswith('<U'):
                    if log_orig.loc[line, param] is np.nan:
                        assert log_orig.loc[line, param] is log_test.loc[line, param]
                    else:
                        assert log_orig.loc[line, param] == log_test.loc[line, param]

                # Float
                else:
                    if param not in ['eqw', 'eqw_err']:
                        assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.05,
                                              equal_nan=True)
                    else:
                        assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.15,
                                              equal_nan=True)

        return

    def test_measurements_xlsx_file(self):

        extension = 'xlsx'
        spec.save_log(baseline_folder/f'test_lines_log.{extension}')

        log_orig = lime.load_log(lines_log_address)
        log_test = lime.load_log(baseline_folder/f'test_lines_log.{extension}')

        for line in spec.log.index:
            for param in spec.log.columns:

                # String
                if _LOG_EXPORT_DICT[param].startswith('<U'):
                    if log_orig.loc[line, param] is np.nan:
                        assert log_orig.loc[line, param] is log_test.loc[line, param]
                    else:
                        assert log_orig.loc[line, param] == log_test.loc[line, param]

                # Float
                else:
                    if param not in ['eqw', 'eqw_err']:
                        assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.05,
                                              equal_nan=True)
                    else:
                        assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.15,
                                              equal_nan=True)

        return

    def test_extra_pages_xlsx(self):

        file_xlsx = baseline_folder/'test_lines_log_multi_page.xlsx'

        if file_xlsx.is_file():
            try:
                remove(file_xlsx)
                print(f"File '{file_xlsx}' has been deleted successfully.")
            except OSError as e:
                print(f"Error: {e.strerror}")

        log_orig = lime.load_log(lines_log_address)

        for page in ['LINELOG', 'LINESLOG2', 'LINELOG']:
            spec.save_log(baseline_folder/file_xlsx, page=page)
            log_test = lime.load_log(file_xlsx)

            for line in spec.log.index:
                for param in spec.log.columns:

                    # String
                    if _LOG_EXPORT_DICT[param].startswith('<U'):
                        if log_orig.loc[line, param] is np.nan:
                            assert log_orig.loc[line, param] is log_test.loc[line, param]
                        else:
                            assert log_orig.loc[line, param] == log_test.loc[line, param]

                    # Float
                    else:
                        if param not in ['eqw', 'eqw_err']:
                            assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.05,
                                                  equal_nan=True)
                        else:
                            assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.15,
                                                  equal_nan=True)

        return

    # def test_measurements_asdf_file(self):
    #
    #     extension = 'asdf'
    #     spec.save_log(baseline_folder/f'test_lines_log.{extension}')
    #
    #     log_orig = lime.load_log(lines_log_address)
    #     log_test = lime.load_log(baseline_folder/f'test_lines_log.{extension}')
    #
    #     for line in spec.log.index:
    #         for param in spec.log.columns:
    #
    #             # String
    #             if _LOG_EXPORT_DICT[param].startswith('<U'):
    #                 if log_orig.loc[line, param] is np.nan:
    #                     assert log_orig.loc[line, param] is log_test.loc[line, param]
    #                 else:
    #                     assert log_orig.loc[line, param] == log_test.loc[line, param]
    #
    #             # Float
    #             else:
    #                 if param not in ['eqw', 'eqw_err']:
    #                     print(param, log_orig.loc[line, param], log_test.loc[line, param])
    #                     assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.05,
    #                                           equal_nan=True)
    #                 else:
    #                     assert np.allclose(log_orig.loc[line, param], log_test.loc[line, param], rtol=0.15,
    #                                           equal_nan=True)
    #
    #     return

    def test_save_load_log(self):

        spec0 = lime.Spectrum(wave_array, flux_array, err_array, redshift=redshift, norm_flux=norm_flux,
                             pixel_mask=pixel_mask)

        spec0.load_log(lines_log_address)

        new_log_spectrum = baseline_folder/ 'manga_lines_log_from_spectrum.txt'
        spec0.save_log(new_log_spectrum)

        assert new_log_spectrum.is_file()

        return


import numpy as np
from functools import partial
from collections.abc import Sequence
from . import __version__

try:
    import pylatex
    pylatex_check = True
except ImportError:
    pylatex_check = False


def numberStringFormat(value, cifras=4):

    if value > 0.001:
        newFormat = f'{value:.{cifras}f}'
    else:
        newFormat = f'{value:.{cifras}e}'

    return newFormat


def format_for_table(entry, rounddig=4, rounddig_er=2, scientific_notation=False, nan_format='none'):

    if rounddig_er == None: #TODO declare a universal tool
        rounddig_er = rounddig

    # Check None entry
    if entry != None:

        # Check string entry
        if isinstance(entry, (str, bytes)):
            formatted_entry = entry

        elif isinstance(entry, (pylatex.MultiColumn, pylatex.MultiRow, pylatex.utils.NoEscape)):
            formatted_entry = entry

        # Case of Numerical entry
        else:

            # Case of an array
            scalarVariable = True
            if isinstance(entry, (Sequence, np.ndarray)):

                # Confirm is not a single value array
                if len(entry) == 1:
                    entry = entry[0]
                # Case of an array
                else:
                    scalarVariable = False
                    formatted_entry = '_'.join(entry)  # we just put all together in a "_' joined string

            # Case single scalar
            if scalarVariable:

                # Case with error quantified # TODO add uncertainty protocol for table
                # if isinstance(entry, UFloat):
                #     formatted_entry = round_sig(nominal_values(entry), rounddig,
                #                                 scien_notation=scientific_notation) + r'$\pm$' + round_sig(
                #         std_devs(entry), rounddig_er, scien_notation=scientific_notation)

                # Case single float
                if np.isnan(entry):
                    formatted_entry = nan_format

                # Case single float
                else:
                    formatted_entry = numberStringFormat(entry, rounddig)
    else:
        # None entry is converted to None
        formatted_entry = 'None'

    return formatted_entry


def table_fluxes(lines_df, table_address, header_format_latex, table_type='pdf', lines_notation=None, store_version=True):

    # Check pylatex is installed else leave
    if pylatex_check:
        pass
    else:
        print(f'\n- WARNING: pylatex is not installed. The table at {table_address} could not be generated')
        return

    # Establish the headers for the table
    n_columns = lines_df.columns.size
    columns_format_list = ['Line'] + ['None'] * n_columns
    for i, column in enumerate(lines_df.columns.values):
        columns_format_list[i + 1] = header_format_latex[column]

    if 'Components' in columns_format_list:
        idx_blended_label = columns_format_list.index('Components')
    else:
        idx_blended_label = None

    # Create pdf
    pdf = PdfMaker()
    pdf.create_pdfDoc(pdf_type='table')
    pdf.pdf_insert_table(columns_format_list)

    # Loop through the lines
    obsLines = lines_df.index.values
    for i, lineLabel in enumerate(obsLines):
        row_raw = [lines_notation[i]] + list(lines_df.loc[lineLabel].values)

        # Exclude the _ from the blended label
        if idx_blended_label is not None:
            row_raw[idx_blended_label] = pylatex.utils.escape_latex(row_raw[idx_blended_label])

        # Add row to the table
        lastRow_check = True if lineLabel == obsLines[-1] else False
        pdf.addTableRow(row_raw, last_row=lastRow_check)

    if store_version:
        versioning = r'\textsc{{{}}} {}'.format(f'LiMe', __version__)
        pdf.table.add_row([pylatex.NoEscape(r'\footnotesize{{{}}}'.format(versioning))],
                          escape=False, strict=False)

    # Save the pdf table
    if table_type == 'pdf':
        try:
            pdf.generate_pdf(table_address, clean_tex=True)
        except:
            print('-- PDF compilation failure')

    return


class PdfMaker:

    def __init__(self):
        """

        """
        self.pdf_type = None
        self.pdf_geometry_options = {'right': '1cm',
                                     'left': '1cm',
                                     'top': '1cm',
                                     'bottom': '2cm'}
        self.table = None
        self.theme_table = None

        # TODO add dictionary with numeric formats for tables depending on the variable

    def create_pdfDoc(self, pdf_type=None, geometry_options=None, document_class=u'article', theme='white'):
        """

        :param pdf_type:
        :param geometry_options:
        :param document_class:
        :param theme:
        :return:
        """
        # TODO integrate this into the init
        # Case for a complete .pdf or .tex
        self.theme_table = theme

        if pdf_type is not None:

            self.pdf_type = pdf_type

            # Update the geometry if necessary (we coud define a dictionary distinction)
            if pdf_type == 'graphs':
                pdf_format = {'landscape': 'true'}
                self.pdf_geometry_options.update(pdf_format)

            elif pdf_type == 'table':
                pdf_format = {'landscape': 'true',
                              'paperwidth': '30in',
                              'paperheight': '30in'}
                self.pdf_geometry_options.update(pdf_format)

            if geometry_options is not None:
                self.pdf_geometry_options.update(geometry_options)

            # Generate the doc
            self.pdfDoc = pylatex.Document(documentclass=document_class, geometry_options=self.pdf_geometry_options)

            if theme == 'dark':
                self.pdfDoc.append(pylatex.NoEscape('\definecolor{background}{rgb}{0.169, 0.169, 0.169}'))
                self.pdfDoc.append(pylatex.NoEscape('\definecolor{foreground}{rgb}{0.702, 0.780, 0.847}'))
                self.pdfDoc.append(pylatex.NoEscape(r'\arrayrulecolor{foreground}'))

            if pdf_type == 'table':
                self.pdfDoc.packages.append(pylatex.Package('preview', options=['active', 'tightpage', ]))
                self.pdfDoc.packages.append(pylatex.Package('hyperref', options=['unicode=true', ]))
                self.pdfDoc.append(pylatex.NoEscape(r'\pagenumbering{gobble}'))
                self.pdfDoc.packages.append(pylatex.Package('nicefrac'))
                self.pdfDoc.packages.append(pylatex.Package('siunitx'))
                self.pdfDoc.packages.append(pylatex.Package('makecell'))
                self.pdfDoc.packages.append(pylatex.Package('footnote'))
                self.pdfDoc.packages.append(pylatex.Package('colortbl', options=['usenames', 'dvipsnames', ]))  # Package to crop pdf to a figure
                self.pdfDoc.packages.append(pylatex.Package('xcolor', options=['table']))

            elif pdf_type == 'longtable':
                self.pdfDoc.append(pylatex.NoEscape(r'\pagenumbering{gobble}'))

        return

    def pdf_create_section(self, caption, add_page=False):

        with self.pdfDoc.create(pylatex.Section(caption)):
            if add_page:
                self.pdfDoc.append(pylatex.NewPage())

    def add_page(self):

        self.pdfDoc.append(pylatex.NewPage())

        return

    def pdf_insert_image(self, image_address, fig_loc='htbp', width=r'1\textwidth'):

        with self.pdfDoc.create(pylatex.Figure(position='h!')) as fig_pdf:
            fig_pdf.add_image(image_address, pylatex.NoEscape(width))

        return

    def pdf_insert_table(self, column_headers=None, table_format=None, addfinalLine=True, color_font=None,
                         color_background=None):

        # Set the table format
        if table_format is None:
            table_format = 'l' + 'c' * (len(column_headers) - 1)

        # Case we want to insert the table in a pdf
        if self.pdf_type != None:

            if self.pdf_type == 'table':
                self.pdfDoc.append(pylatex.NoEscape(r'\begin{preview}'))

                # Initiate the table
                with self.pdfDoc.create(pylatex.Tabular(table_format)) as self.table:
                    if column_headers != None:
                        self.table.add_hline()
                        # self.table.add_row(list(map(str, column_headers)), escape=False, strict=False)
                        output_row = list(map(partial(format_for_table), column_headers))

                        # if color_font is not None:
                        #     for i, item in enumerate(output_row):
                        #         output_row[i] = NoEscape(r'\color{{{}}}{}'.format(color_font, item))
                        #
                        # if color_background is not None:
                        #     for i, item in enumerate(output_row):
                        #         output_row[i] = NoEscape(r'\cellcolor{{{}}}{}'.format(color_background, item))

                        if (color_font is not None) or (self.theme_table != 'white'):
                            if self.theme_table == 'dark' and color_font is None:
                                color_font = 'foreground'

                            for i, item in enumerate(output_row):
                                output_row[i] = pylatex.NoEscape(r'\color{{{}}}{}'.format(color_font, item))

                        if (color_background is not None) or (self.theme_table != 'white'):
                            if self.theme_table == 'dark' and color_background is None:
                                color_background = 'background'

                            for i, item in enumerate(output_row):
                                output_row[i] = pylatex.NoEscape(r'\cellcolor{{{}}}{}'.format(color_background, item))

                        self.table.add_row(output_row, escape=False, strict=False)
                        if addfinalLine:
                            self.table.add_hline()

            elif self.pdf_type == 'longtable':

                # Initiate the table
                with self.pdfDoc.create(pylatex.LongTable(table_format)) as self.table:
                    if column_headers != None:
                        self.table.add_hline()
                        self.table.add_row(list(map(str, column_headers)), escape=False)
                        if addfinalLine:
                            self.table.add_hline()

        # Table .tex without preamble
        else:
            self.table = pylatex.Tabu(table_format)
            if column_headers != None:
                self.table.add_hline()
                # self.table.add_row(list(map(str, column_headers)), escape=False, strict=False)
                output_row = list(map(partial(format_for_table), column_headers))
                self.table.add_row(output_row, escape=False, strict=False)
                if addfinalLine:
                    self.table.add_hline()

        return

    def pdf_insert_longtable(self, column_headers=None, table_format=None):

        # Set the table format
        if table_format is None:
            table_format = 'l' + 'c' * (len(column_headers) - 1)

        # Case we want to insert the table in a pdf
        if self.pdf_type != None:

            if self.pdf_type == 'table':
                self.pdfDoc.append(pylatex.NoEscape(r'\begin{preview}'))

                # Initiate the table
            with self.pdfDoc.create(pylatex.Tabu(table_format)) as self.table:
                if column_headers != None:
                    self.table.add_hline()
                    self.table.add_row(map(str, column_headers), escape=False)
                    self.table.add_hline()

                    # Table .tex without preamble
        else:
            self.table = pylatex.LongTable(table_format)
            if column_headers != None:
                self.table.add_hline()
                self.table.add_row(list(map(str, column_headers)), escape=False)
                self.table.add_hline()

    def addTableRow(self, input_row, row_format='auto', rounddig=4, rounddig_er=None, last_row=False, color_font=None,
                    color_background=None):

        # Default formatting
        if row_format == 'auto':
            output_row = list(map(partial(format_for_table, rounddig=rounddig), input_row))
        else:
            print('BICHO')
            output_row = input_row

        # TODO clean this theming to default values
        if (color_font is not None) or (self.theme_table != 'white'):
            if self.theme_table == 'dark' and color_font is None:
                color_font = 'foreground'

            for i, item in enumerate(output_row):
                output_row[i] = pylatex.NoEscape(r'\color{{{}}}{}'.format(color_font, item))

        if (color_background is not None) or (self.theme_table != 'white'):
            if self.theme_table == 'dark' and color_background is None:
                color_background = 'background'

            for i, item in enumerate(output_row):
                output_row[i] = pylatex.NoEscape(r'\cellcolor{{{}}}{}'.format(color_background, item))

        # Append the row
        self.table.add_row(output_row, escape=False, strict=False)

        # Case of the final row just add one line
        if last_row:
            self.table.add_hline()

    def fig_to_pdf(self, label=None, fig_loc='htbp', width=r'1\textwidth', add_page=False, *args, **kwargs):

        with self.pdfDoc.create(pylatex.Figure(position=fig_loc)) as plot:
            plot.add_plot(width=pylatex.NoEscape(width), placement='h', *args, **kwargs)

            if label is not None:
                plot.add_caption(label)

        if add_page:
            self.pdfDoc.append(pylatex.NewPage())

    def generate_pdf(self, output_address, clean_tex=True):

        if self.pdf_type is None:
            self.table.generate_tex(str(output_address))

        else:
            if self.pdf_type == 'table':
                self.pdfDoc.append(pylatex.NoEscape(r'\end{preview}'))
            self.pdfDoc.generate_pdf(filepath=str(output_address), clean_tex=clean_tex, compiler='pdflatex')

        return





from pathlib import Path
import numpy as np
import lime

conf_file_address = Path(__file__).parent/'data_tests'/'manga.toml'
lines_log_address = Path(__file__).parent/'data_tests'/'manga_lines_log.txt'

cfg = lime.load_cfg(conf_file_address)

def test_load_cfg():

    assert 'SHOC579' in cfg
    assert 'default_line_fitting' in cfg
    assert 'MASK_0_line_fitting' in cfg
    assert '38-35_line_fitting' in cfg

    assert 'H1_4861A_k-1_sigma' in cfg['MASK_0_line_fitting']
    assert isinstance(cfg['MASK_0_line_fitting']['H1_4861A_k-1_sigma'], dict)
    assert cfg['MASK_0_line_fitting']['H1_4861A_k-1_sigma']['expr'] == '>2.0*H1_4861A_sigma'

    assert 'H1_9548A_sigma' in cfg['MASK_0_line_fitting']
    assert isinstance(cfg['MASK_0_line_fitting']['H1_9548A_sigma'], dict)
    assert cfg['MASK_0_line_fitting']['H1_9548A_sigma']['min'] == 1.0
    assert cfg['MASK_0_line_fitting']['H1_9548A_sigma']['max'] == 2.0

    return


def test_save_cfg():

    save_file_address = Path(__file__).parent/'data_tests'/'new_manga.toml'
    copy_cfg = cfg.copy()
    copy_cfg['new_line_fitting'] = {'O3_5007A_kinem': "O3_4959A",
                                    'O3_5007A_k-1_kinem': "O3_4959A_k-1",
                                    'He1_5016A_center': "min:5014,max:5018",
                                    'He1_5016A_sigma': "min:1.0,max:2.0"
                                    }

    lime.save_cfg(copy_cfg, save_file_address)

    new_cfg = lime.load_cfg(save_file_address)

    assert 'SHOC579' in new_cfg
    assert 'default_line_fitting' in new_cfg
    assert 'MASK_0_line_fitting' in new_cfg
    assert '38-35_line_fitting' in new_cfg

    assert 'H1_4861A_k-1_sigma' in new_cfg['MASK_0_line_fitting']
    assert isinstance(new_cfg['MASK_0_line_fitting']['H1_4861A_k-1_sigma'], dict)
    assert new_cfg['MASK_0_line_fitting']['H1_4861A_k-1_sigma']['expr'] == '>2.0*H1_4861A_sigma'

    assert 'H1_9548A_sigma' in new_cfg['MASK_0_line_fitting']
    assert isinstance(new_cfg['MASK_0_line_fitting']['H1_9548A_sigma'], dict)
    assert new_cfg['MASK_0_line_fitting']['H1_9548A_sigma']['min'] == 1.0
    assert new_cfg['MASK_0_line_fitting']['H1_9548A_sigma']['max'] == 2.0

    assert 'new_line_fitting' in new_cfg
    assert isinstance(new_cfg['new_line_fitting']['He1_5016A_center'], dict)
    assert isinstance(new_cfg['new_line_fitting']['He1_5016A_sigma'], dict)
    assert new_cfg['new_line_fitting']['He1_5016A_center']['min'] == 5014.0
    assert new_cfg['new_line_fitting']['He1_5016A_center']['max'] == 5018.0
    assert new_cfg['new_line_fitting']['He1_5016A_sigma']['min'] == 1.0
    assert new_cfg['new_line_fitting']['He1_5016A_sigma']['max'] == 2.0

    return


def test_log_parameters_calculation():

    log_lines = lime.load_log(lines_log_address)
    lime.extract_fluxes(log_lines, flux_type='mixture')

    parameters = ['eqw_new',
                  'eqw_new_err']

    formulation = ['line_flux/cont',
                   '(line_flux/cont) * sqrt((line_flux_err/gauss_flux)**2 + (std_cont/cont)**2)']

    lime.log_parameters_calculation(log_lines, parameters, formulation)
    assert 'eqw_new' in log_lines.columns
    tol = log_lines['eqw_err'] * 2
    assert np.allclose(log_lines['eqw_new'], log_lines['eqw'], atol=tol)

    return